import functools
import os
import shutil
from typing import (
    List,
    Optional,
    Tuple,
)

from typing_extensions import Unpack

from galaxy import exceptions
from galaxy.files import OptionalUserContext
from galaxy.util.path import (
    safe_contains,
    safe_path,
    safe_walk,
)
from . import (
    AnyRemoteEntry,
    BaseFilesSource,
    FilesSourceOptions,
    FilesSourceProperties,
)

DEFAULT_ENFORCE_SYMLINK_SECURITY = True
DEFAULT_DELETE_ON_REALIZE = False
DEFAULT_ALLOW_SUBDIR_CREATION = True
DEFAULT_PREFER_LINKS = False


class PosixFilesSourceProperties(FilesSourceProperties, total=False):
    root: str
    enforce_symlink_security: bool
    delete_on_realize: bool
    allow_subdir_creation: bool
    prefer_links: bool


class PosixFilesSource(BaseFilesSource):
    plugin_type = "posix"

    # If this were a PyFilesystem2FilesSource all that would be needed would be,
    # but we couldn't enforce security our way I suspect.
    # def _open_fs(self):
    #    from fs.osfs import OSFS
    #    handle = OSFS(**self._props)
    #    return handle

    def __init__(self, **kwd: Unpack[PosixFilesSourceProperties]):
        props = self._parse_common_config_opts(kwd)
        self.root = props.get("root")
        if not self.root:
            self.writable = False
        self.enforce_symlink_security = props.get("enforce_symlink_security", DEFAULT_ENFORCE_SYMLINK_SECURITY)
        self.delete_on_realize = props.get("delete_on_realize", DEFAULT_DELETE_ON_REALIZE)
        self.allow_subdir_creation = props.get("allow_subdir_creation", DEFAULT_ALLOW_SUBDIR_CREATION)
        self._prefer_links = props.get("prefer_links", DEFAULT_PREFER_LINKS)

    def prefer_links(self) -> bool:
        return self._prefer_links

    def _list(
        self,
        path="/",
        recursive=True,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        sort_by: Optional[str] = None,
    ) -> Tuple[List[AnyRemoteEntry], int]:
        if not self.root:
            raise exceptions.ItemAccessibilityException("Listing files at file:// URLs has been disabled.")
        dir_path = self._to_native_path(path, user_context=user_context)
        if not self._safe_directory(dir_path):
            raise exceptions.ObjectNotFound(f"The specified directory does not exist [{dir_path}].")
        if recursive:
            res: List[AnyRemoteEntry] = []
            effective_root = self._effective_root(user_context)
            for p, dirs, files in safe_walk(dir_path, allowlist=self._allowlist):
                rel_dir = os.path.relpath(p, effective_root)
                to_dict = functools.partial(self._resource_info_to_dict, rel_dir, user_context=user_context)
                res.extend(map(to_dict, dirs))
                res.extend(map(to_dict, files))
            return res, len(res)
        else:
            res = os.listdir(dir_path)
            to_dict = functools.partial(self._resource_info_to_dict, path, user_context=user_context)
            return list(map(to_dict, res)), len(res)

    def _realize_to(
        self,
        source_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        if not self.root and (not user_context or not user_context.is_admin):
            raise exceptions.ItemAccessibilityException("Writing to file:// URLs has been disabled.")

        effective_root = self._effective_root(user_context)
        source_native_path = self._to_native_path(source_path, user_context=user_context)
        if self.enforce_symlink_security:
            if not safe_contains(effective_root, source_native_path, allowlist=self._allowlist):
                raise Exception("Operation not allowed.")
        else:
            source_native_path = os.path.normpath(source_native_path)
            assert source_native_path.startswith(os.path.normpath(effective_root))

        if not self.delete_on_realize:
            shutil.copyfile(source_native_path, native_path)
        else:
            shutil.move(source_native_path, native_path)

    def _write_from(
        self,
        target_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        effective_root = self._effective_root(user_context)
        target_native_path = self._to_native_path(target_path, user_context=user_context)
        if self.enforce_symlink_security:
            if not safe_contains(effective_root, target_native_path, allowlist=self._allowlist):
                raise Exception("Operation not allowed.")
        else:
            target_native_path = os.path.normpath(target_native_path)
            assert target_native_path.startswith(os.path.normpath(effective_root))

        target_native_path_parent, target_native_path_name = os.path.split(target_native_path)
        if not os.path.exists(target_native_path_parent):
            if self.allow_subdir_creation:
                os.makedirs(target_native_path_parent)
            else:
                raise Exception("Parent directory does not exist.")

        # Use a temporary name while writing so anything that consumes written files can detect when they've completed,
        # and identify interrupted writes
        target_native_path_part = os.path.join(target_native_path_parent, f"_{target_native_path_name}.part")
        shutil.copyfile(native_path, target_native_path_part)
        os.rename(target_native_path_part, target_native_path)

    def _to_native_path(self, source_path: str, user_context: OptionalUserContext = None):
        source_path = os.path.normpath(source_path)
        if source_path.startswith("/"):
            source_path = source_path[1:]
        return os.path.join(self._effective_root(user_context), source_path)

    def _effective_root(self, user_context: OptionalUserContext = None):
        return self._evaluate_prop(self.root or "/", user_context=user_context)

    def _resource_info_to_dict(self, dir: str, name: str, user_context: OptionalUserContext = None) -> AnyRemoteEntry:
        rel_path = os.path.normpath(os.path.join(dir, name))
        full_path = self._to_native_path(rel_path, user_context=user_context)
        uri = self.uri_from_path(rel_path)
        if os.path.isdir(full_path):
            return {"class": "Directory", "name": name, "uri": uri, "path": rel_path}
        else:
            statinfo = os.lstat(full_path)
            return {
                "class": "File",
                "name": name,
                "size": statinfo.st_size,
                "ctime": self.to_dict_time(statinfo.st_ctime),
                "uri": uri,
                "path": rel_path,
            }

    def _safe_directory(self, directory):
        if self.enforce_symlink_security:
            if not safe_path(directory, allowlist=self._allowlist):
                raise exceptions.ConfigDoesNotAllowException(
                    f"directory ({directory}) is a symlink to a location not on the allowlist"
                )

        if not os.path.exists(directory):
            return False
        return True

    def _serialization_props(self, user_context: OptionalUserContext = None) -> PosixFilesSourceProperties:
        return {
            # abspath needed because will be used by external Python from
            # a job working directory
            "root": os.path.abspath(self._effective_root(user_context)),
            "enforce_symlink_security": self.enforce_symlink_security,
            "delete_on_realize": self.delete_on_realize,
            "allow_subdir_creation": self.allow_subdir_creation,
            "prefer_links": self._prefer_links,
        }

    @property
    def _allowlist(self):
        return self._file_sources_config.symlink_allowlist

    def score_url_match(self, url: str):
        # For security, we need to ensure that a partial match doesn't work. e.g. file://{root}something/myfiles
        if self.root and (
            url.startswith(f"{self.get_uri_root()}://{self.root}/") or url == f"self.get_uri_root()://{self.root}"
        ):
            return len(f"self.get_uri_root()://{self.root}")
        elif self.root and (url.startswith(f"file://{self.root}/") or url == f"file://{self.root}"):
            return len(f"file://{self.root}")
        elif not self.root and url.startswith("file://"):
            return len("file://")
        else:
            return super().score_url_match(url)

    def to_relative_path(self, url: str) -> str:
        if url.startswith(f"file://{self.root}"):
            return url[len(f"file://{self.root}") :]
        elif url.startswith("file://"):
            return url[7:]
        else:
            return super().to_relative_path(url)


__all__ = ("PosixFilesSource",)



from typing import Union

try:
    from fs.azblob import (
        BlobFS,
        BlobFSV2,
    )
except ImportError:
    BlobFS = None

from typing import Optional

from . import (
    FilesSourceOptions,
    FilesSourceProperties,
)
from ._pyfilesystem2 import PyFilesystem2FilesSource


class AzureFileSource(PyFilesystem2FilesSource):
    plugin_type = "azure"
    required_module = BlobFS
    required_package = "fs-azureblob"

    def _open_fs(self, user_context=None, opts: Optional[FilesSourceOptions] = None):
        props = self._serialization_props(user_context)
        extra_props: Union[FilesSourceProperties, dict] = opts.extra_props or {} if opts else {}
        all_props = {**props, **extra_props}
        namespace_type = all_props.get("namespace_type", "hierarchical")
        if namespace_type not in ["hierarchical", "flat"]:
            raise Exception("Misconfigured azure file source")
        account_name = all_props["account_name"]
        account_key = all_props["account_key"]
        container = all_props["container_name"]
        handle: Union[BlobFSV2, BlobFS]
        if namespace_type == "flat":
            handle = BlobFS(account_name, container, account_key)
        else:
            handle = BlobFSV2(account_name, container, account_key)

        return handle


__all__ = ("AzureFileSource",)



"""
Galaxy FilesSource implementation for eLabFTW.

This module implements a FilesSource that interacts with an eLabFTW [1] instance. eLabFTW revolves around the concepts
of *experiment* [2] and *resource* [3]. Experiments and resources can have files attached to them. To get a quick
overview, try out the live demo [4]. The scope of this implementation is exporting data from and importing data to
eLabFTW as file attachments of *already existing* experiments and resources. Each user can configure their preferred
eLabFTW instance entering its URL and an API Key.
File sources reference files via a URI, while eLabFTW uses auto-incrementing positive integers. For more details read
galaxyproject/galaxy#18665 [5]. This leads to the need to declare a mapping between said identifiers and Galaxy URIs.

Those take the form ``elabftw://demo.elabftw.net/entity_type/entity_id/attachment_id``, where:

- ``entity_type`` is either 'experiments' or 'resources'
- ``entity_id`` is the id (an integer in string form) of an experiment or resource
- ``attachment_id`` is the id (an integer in string form) of an attachment

For the user-defined file sources use case (when users configure one or more instances of the FilesSource via a file
source template), Galaxy URIs have a different scheme and authority, taking the form ``gxuserfiles://file_source_id/
entity_type/entity_id/attachment_id``, where:

- ``file_source_id`` is the file source identifier assigned by Galaxy

This implementation uses both ``aiohttp`` and the ``requests`` libraries as underlying mechanisms to communicate with
eLabFTW via its REST API [6]. A significant limitation of the implementation is that, due to the fact that the API does
not have an endpoint that can list attachments for several experiments and/or resources with a single request, when
listing the root directory or an entity type *recursively*, a list of entities has to be fetched first, then to fetch
the information on their attachments, a separate request has to be sent *for each one* of them. The ``aiohttp`` library
makes it bearable to recursively browse instances with up to ~500 experiments or resources with attachments by sending
them concurrently, but ultimately solving the problem would require changes to the API from the eLabFTW side. Another
limitation is that it does not support pagination, since eLabFTW does not return the total count of experiments or
resources matching a query, therefore the amount of pages that should be displayed on the Galaxy user interface is
unknown.

References:

- [1] https://www.elabftw.net/
- [2] https://doc.elabftw.net/user-guide.html#experiments
- [3] https://doc.elabftw.net/user-guide.html#resources
- [4] https://demo.elabftw.net
- [5] https://github.com/galaxyproject/galaxy/issues/18665
- [6] https://doc.elabftw.net/api/v2
"""

import asyncio
import logging
import re
from abc import ABC
from datetime import (
    datetime,
    timezone,
)
from pathlib import Path
from textwrap import dedent
from time import time
from typing import (
    Any,
    AsyncIterator,
    cast,
    Dict,
    Generic,
    get_type_hints,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
)
from urllib.parse import (
    ParseResult,
    urljoin,
    urlparse,
)

import aiohttp
from requests import Session as RequestsSession
from typing_extensions import (
    NotRequired,
    TypedDict,
    Unpack,
)

from galaxy import exceptions as galaxy_exceptions
from galaxy.files import OptionalUserContext
from galaxy.files.sources import (
    AnyRemoteEntry,
    BaseFilesSource,
    DEFAULT_SCHEME,
    FilesSourceOptions,
    FilesSourceProperties,
    PluginKind,
    RemoteDirectory,
    RemoteFile,
)
from galaxy.util import requests

__all__ = ("eLabFTWFilesSource",)

MAX_CONCURRENT_REQUESTS: int = 75  # max concurrent requests to eLabFTW (e.g. recursively listing a directory)
MAX_ITEMS_PER_PAGE: int = 1000  # max items per page when requesting experiments or resources
CONNECT_TIMEOUT: int = 10  # time out connections to eLabFTW after this number of seconds
READ_TIMEOUT: int = 10  # time out waiting for responses from eLabFTW after this number of seconds
PAGINATION_TIMEOUT: int = 30  # stop processing paginated responses after this number of seconds


eLabFTWRemoteEntryWrapperType = TypeVar("eLabFTWRemoteEntryWrapperType", bound=AnyRemoteEntry)


class eLabFTWRemoteEntryWrapper(Generic[eLabFTWRemoteEntryWrapperType]):  # noqa
    """
    Wrap a remote entry produced by this module to easily access its entity type, entity id, and attachment id.
    """

    def __init__(self, entry: eLabFTWRemoteEntryWrapperType, source: Optional[dict] = None):
        """
        Initialize the remote entry wrapper.

        :param entry: Remote entry to be wrapped.
        :type entry: eLabFTWRemoteEntryWrapperType
        :param source: Information used to construct the remote entry.
        :type source: Optional[dict]
        """
        self.entry = entry
        self.source = source

    @property
    def entity_type(self) -> Optional[str]:
        """
        Get the entity type for the wrapped entry.
        """
        return self._get_part("entity_type")

    @property
    def entity_id(self) -> Optional[str]:
        """
        Get the entity id for the wrapped entry.
        """
        return self._get_part("entity_id")

    @property
    def attachment_id(self) -> Optional[str]:
        """
        Get the attachment id for the wrapped entry.
        """
        return self._get_part("attachment_id")

    def _get_part(self, part: Literal["entity_type", "entity_id", "attachment_id"]) -> Optional[str]:
        """
        Get the entity type, entity id or attachment id for the wrapped entry.
        """
        path = self.entry["path"]
        entity_type, entity_id, attachment_id = split_path(path)
        return locals()[part]


class eLabFTWFilesSourceProperties(FilesSourceProperties, total=False):  # noqa
    endpoint: str
    api_key: str


class eLabFTWFilesSource(BaseFilesSource):  # noqa

    plugin_type = "elabftw"
    plugin_kind = PluginKind.rfs
    supports_pagination = False
    # reason: eLabFTW does not return the total count of elements matching a query, see
    # https://github.com/galaxyproject/galaxy/pull/19319#discussion_r1928753352
    supports_search = True
    supports_sorting = True

    def __init__(self, *args, **kwargs: Unpack[eLabFTWFilesSourceProperties]):
        """Initialize the eLabFTW files source with an API key and an endpoint URL."""
        super().__init__()
        props = self._parse_common_config_opts(kwargs)
        self._props = props

        self._endpoint = kwargs["endpoint"]  # meant to be accessed only from `_get_endpoint()`
        self._api_key = kwargs["api_key"]  # meant to be accessed only from `_create_session()`

    def get_prefix(self, user_context: OptionalUserContext = None) -> Optional[str]:
        endpoint: ParseResult = self._get_endpoint(user_context=user_context)
        return self.id if self.scheme not in {"elabftw", DEFAULT_SCHEME} else (endpoint.netloc or None)
        # it would make better sense to return
        # `self.id if self.scheme == USER_FILE_SOURCES_SCHEME else (endpoint.netloc or None)`, where
        # `USER_FILE_SOURCES_SCHEME` comes from `galaxy.managers.file_source_instances`; however, that would lead to a
        # circular import (maybe `USER_FILE_SOURCES_SCHEME` should be moved to a module in a layer deeper than
        # `galaxy.managers`)

    def get_scheme(self) -> str:
        return self.scheme if self.scheme and self.scheme != DEFAULT_SCHEME else "elabftw"
        # it would make better sense to return `self.scheme if self.scheme == USER_FILE_SOURCES_SCHEME else "elabftw"`,
        # but the same circular import issue as above arises

    def score_url_match(self, url: str) -> int:
        parsed_url = urlparse(url)
        return sum(
            int(check)
            for check in (
                parsed_url.scheme == self.get_scheme(),
                parsed_url.netloc == self.get_prefix(),
            )
        )

    def to_relative_path(self, url: str) -> str:
        parsed_url = urlparse(url)
        path = parsed_url.path
        if not path.startswith("/"):
            path = f"/{path}"
        return path

    def _create_session(
        self,
        options: Optional[FilesSourceOptions] = None,
        user_context: OptionalUserContext = None,
    ) -> RequestsSession:
        """
        Create a Galaxy ``requests`` session, overriding initial settings via a :class:`FileSourceOptions` object.
        """
        return requests.Session(
            headers=self._get_session_headers(options=options, user_context=user_context),  # type: ignore[call-arg]
        )

    def _create_session_async(
        self,
        options: Optional[FilesSourceOptions] = None,
        user_context: OptionalUserContext = None,
    ) -> aiohttp.ClientSession:
        """
        Create an ``aiohttp`` session, overriding initial settings via a :class:`FileSourceOptions` object.
        """
        connector = aiohttp.TCPConnector(limit=MAX_CONCURRENT_REQUESTS)
        return aiohttp.ClientSession(
            connector=connector,
            raise_for_status=True,
            headers=self._get_session_headers(options=options, user_context=user_context),
        )

    def _get_session_headers(
        self,
        options: Optional[FilesSourceOptions] = None,
        user_context: OptionalUserContext = None,
    ) -> dict:
        """
        Construct a dictionary of HTTP client session headers.

        Optionally, override initial settings via a :class:`FileSourceOptions` object and/or a
        :class:`FileSourcesUserContext` object.

        Meant to be used only by `_create_session()` and `_create_session_async()`.
        """
        props = {}
        props.update(self._props)
        props.update(options.extra_props if options and options.extra_props else {})
        props.update(
            {key: value for key, value in self._serialization_props(user_context).items() if value is not None}
        )
        headers = {
            "Authorization": props.get("api_key", self._api_key),
            "Accept": "application/json",
        }
        return headers

    def _get_endpoint(
        self,
        options: Optional[FilesSourceOptions] = None,
        user_context: OptionalUserContext = None,
    ) -> ParseResult:
        """
        Retrieve the endpoint from the constructor, or override it via a :class:`FileSourceOptions` object.
        """
        props = {}
        props.update(self._props)
        props.update(options.extra_props if options and options.extra_props else {})
        props.update(
            {key: value for key, value in self._serialization_props(user_context).items() if value is not None}
        )
        endpoint = props.get("endpoint", self._endpoint)
        # given that `options.extra_props` is of `eLabFTWFilesSourceProperties` type, it should be a string
        endpoint = cast(str, endpoint)

        return urlparse(endpoint)

    def _serialization_props(self, user_context: OptionalUserContext = None) -> eLabFTWFilesSourceProperties:
        effective_props: Dict[str, Any] = {}

        for key, val in self._props.items():
            if key in {"api_key", "endpoint"} and user_context is None:
                # prevent exception while expanding `${user.user_vault.read_secret('preferences/elabftw/api_key')}` or
                # `${user.preferences['elabftw|endpoint']}` without `user_context`
                effective_props[key] = None
                continue
            effective_props[key] = self._evaluate_prop(val, user_context=user_context)

        return cast(eLabFTWFilesSourceProperties, effective_props)

    def _list(
        self,
        path="/",
        recursive=False,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        sort_by: Optional[str] = None,
        # in particular, expecting
        # `sort_by: Optional[Literal["name", "uri", "path", "class", "size", "ctime"]] = None,`
        # from Python 3.9 on, the following would be possible, although barely readable
        # `sort_by: Optional[Literal[*(get_type_hints(RemoteDirectory) | get_type_hints(RemoteFile)).keys()]] = None,`
    ) -> Tuple[List[AnyRemoteEntry], int]:
        """
        List the contents of an eLabFTW endpoint.

        FastAPI runs sync routes on a separate thread (from a threadpool), and async routes on the main thread's event
        loop. Originally, it was conceived to convert `_list` to an async method (see
        https://github.com/galaxyproject/galaxy/pull/19256). However, given that all other file source plugins are
        blocking, they could block the main thread for a significant amount of time (see
        https://github.com/fastapi/fastapi/issues/3091). Thus, the implementation below creates a new event loop within
        the sync route's thread so that the eLabFTW plugin can still send concurrent requests without blocking the main
        thread.
        """
        event_loop = asyncio.new_event_loop()
        try:
            asyncio.set_event_loop(event_loop)
            return event_loop.run_until_complete(
                self._list_async(
                    path=path,
                    recursive=recursive,
                    user_context=user_context,
                    opts=opts,
                    limit=limit,
                    offset=offset,
                    query=query,
                    sort_by=sort_by,
                ),
            )
        finally:
            event_loop.close()

    async def _list_async(
        self,
        path="/",
        recursive=False,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        sort_by: Optional[str] = None,
        # in particular, expecting
        # `sort_by: Optional[Literal["name", "uri", "path", "class", "size", "ctime"]] = None,`
    ) -> Tuple[List[AnyRemoteEntry], int]:
        """
        List remote entries in a remote directory.

        List entity types ("experiment" and "resource"), entity ids of a specific type, or the ids of files attached to
        an entity.

        Meant to be called only by `_list()`.

        :param path: Path referring to the root, an entity type, or an entity id.
        :type path: str
        :param recursive: List recursively, including all entity types for the root, all entities for each entity type
                          and all attachments for each entity.
        :type recursive: bool
        :param user_context: Alter behavior using information from a user context (e.g. override the API key).
        :type user_context: OptionalUserContext
        :param opts: Alter behavior using information from a file source options object (e.g. ignore locked resources).
        :type opts: Optional[FilesSourceOptions]
        :param limit: Show at most this amount of results, defaults to unlimited.
        :type limit: Optional[int]
        :param offset: Filter out this amount of results from the beginning of the sequence, defaults to zero.
        :type offset: Optional[int]
        :param query: Show only results that contain this string.
        :type query: Optional[int]
        :param sort_by: Sort results by name, URI, path, class (directory or file), size (files only) or creation time
                        (files only).
        :type sort_by: Optional[str]

        :raises aiohttp.ClientError: When there is a connection error.
        :raises ValidationError: If any HTTP response from the eLabFTW server is invalid.
        :raises DirectoryExpected: If the path refers to an attachment id.
        :raises InvalidPath: Path constraints described in the docstring of :class:`InvalidPath` are not satisfied.
        :raises ResourceNotFound: If the path refers to a non-existing experiment, resource, or attachment.
        """
        session = self._create_session_async(options=opts, user_context=user_context)
        endpoint = self._get_endpoint(options=opts, user_context=user_context)

        entity_type, entity_id, attachment_id = split_path(path)

        async with session:
            retrieve_entity_types: bool = not entity_type
            retrieve_entities: bool = bool((entity_type and not entity_id) or (not entity_id and recursive))
            retrieve_entities_server_side_offset = offset if retrieve_entities and not recursive else None
            retrieve_attachments: bool = bool((entity_type and entity_id) or (not attachment_id and recursive))

            async def regular_iterable_to_async_iterator(regular_iter: Iterable) -> AsyncIterator:
                """
                Convert a regular iterable to an async iterator.
                """
                for item in regular_iter:
                    yield item

            async def collect_async_iterator(async_iter: AsyncIterator) -> list:
                """
                Collect values of an async iterator into a list.
                """
                return [value async for value in async_iter]

            fetch_entity_types_tasks: List[asyncio.Task] = (
                # fmt: off
                [
                    asyncio.create_task(
                        collect_async_iterator(
                            self._yield_entity_types(
                                endpoint,
                                session,
                                user_context=user_context,
                            )
                        )
                    )
                ]
                # fmt: on
                if retrieve_entity_types
                else []
            )
            fetch_entities_tasks: List[asyncio.Task] = (
                [
                    asyncio.create_task(
                        collect_async_iterator(
                            self._yield_entities(
                                (
                                    # both `wrapped_entity_type.entity_type` and `entity_type` should not be `None` if
                                    # they are used
                                    cast(str, wrapped_entity_type.entity_type)
                                    if retrieve_entity_types
                                    else cast(str, entity_type)
                                ),
                                endpoint,
                                session,
                                limit=(
                                    limit
                                    if not (recursive and (sort_by in {"name", "class", "size", "ctime"} or query))
                                    else None
                                ),
                                offset=retrieve_entities_server_side_offset,
                                query=query if not recursive else None,
                                order=(
                                    # map Galaxy `sort_by` parameter to an eLabFTW API query param
                                    {
                                        "name": "title",
                                        "uri": "id",
                                        "path": "id",
                                    }.get(sort_by)
                                    if isinstance(sort_by, str)
                                    else None
                                ),
                                writable=self.writable,
                                user_context=user_context,
                            )
                        )
                    )
                    async for wrapped_entity_type in (
                        (
                            wrapped_entity_type
                            for coroutine in asyncio.as_completed(fetch_entity_types_tasks)
                            for wrapped_entity_type in (await coroutine)
                        )
                        if retrieve_entity_types
                        else regular_iterable_to_async_iterator([None])
                    )
                ]
                if retrieve_entities
                else []
            )
            fetch_attachments_tasks: List[asyncio.Task] = (
                # fetching attachments is "bearable" for the user up to ~500 experiments + resources with attachments;
                # if eLabFTW allowed listing attachments without having to send individual requests for each experiment
                # or resource, this would not be a concern
                [
                    asyncio.create_task(
                        collect_async_iterator(
                            self._yield_attachments(
                                # all of `wrapped_entity_type.entity_type`, `wrapped_entity_type.entity_id`,
                                # `entity_type` and `entity_id` should not be `None` if they are used
                                cast(str, wrapped_entity.entity_type) if retrieve_entities else cast(str, entity_type),
                                cast(str, wrapped_entity.entity_id) if retrieve_entities else cast(str, entity_id),
                                endpoint,
                                session,
                                user_context=user_context,
                            )
                        )
                    )
                    async for wrapped_entity in (
                        (
                            wrapped_entity
                            for coroutine in asyncio.as_completed(fetch_entities_tasks)
                            for wrapped_entity in (await coroutine)
                            if wrapped_entity.source["has_attachment"]
                        )
                        if retrieve_entities
                        else regular_iterable_to_async_iterator([None])
                    )
                ]
                if retrieve_attachments
                else []
            )

            wrapped_entity_types: List[eLabFTWRemoteEntryWrapper[RemoteDirectory]] = [
                wrapped_entity_type
                for wrapped_entity_types in await asyncio.gather(*fetch_entity_types_tasks)
                for wrapped_entity_type in wrapped_entity_types
            ]
            wrapped_entities: List[eLabFTWRemoteEntryWrapper[RemoteDirectory]] = [
                wrapped_entity
                for wrapped_entities in await asyncio.gather(*fetch_entities_tasks)
                for wrapped_entity in wrapped_entities
            ]
            wrapped_attachments: List[eLabFTWRemoteEntryWrapper[RemoteFile]] = [
                wrapped_attachment
                for wrapped_attachments in await asyncio.gather(*fetch_attachments_tasks)
                for wrapped_attachment in wrapped_attachments
            ]

        if attachment_id:
            attachment_ids = {
                wrapped_attachment.source["id"]
                for wrapped_attachment in wrapped_attachments
                if not isinstance(wrapped_attachment.source, type(None))
            }
            if attachment_id in attachment_ids:
                raise DirectoryExpected(err_msg=f"'{path}' is a file, it cannot be listed")
            else:
                raise ResourceNotFound(err_msg=f"'{path} does not exist")

        wrapped_entries = wrapped_entity_types + wrapped_entities + wrapped_attachments

        # results arrive from the server in nondeterministic order; even if `sort_by` is `None`, calling `_list` twice
        # with the same arguments should return the same results in the same order (otherwise the option `offset` makes
        # no sense).
        constructors = {**get_type_hints(RemoteDirectory), **get_type_hints(RemoteFile)}
        wrapped_entries = sorted(
            wrapped_entries,
            key=lambda x: (
                (
                    x.entry.get(sort_by, constructors[sort_by]())  # fall back to the default object for this key type
                    if sort_by is not None else None  # fmt: skip
                ),
                x.entry["uri"],  # ensure deterministic ordering (URIs are unique)
            ),
        )

        # filter out remaining items locally; by `query`, `offset` and `limit`
        if query is not None:
            wrapped_entries = [
                wrapped_entry for wrapped_entry in wrapped_entries if query in wrapped_entry.entry.get("name", "")
            ]
        if offset is not None:
            wrapped_entries = wrapped_entries[offset - (retrieve_entities_server_side_offset or 0) :]
        if limit is not None:
            wrapped_entries = wrapped_entries[:limit]

        # The second return value is meant to be the total count of elements that match the current filter. Since the
        # file source does not support pagination (both `limit` and `offset` are always `None`), the length of `entries`
        # always matches such value.
        return (entries := [wrapped_entry.entry for wrapped_entry in wrapped_entries]), len(entries)

    async def _yield_entity_types(
        self,
        endpoint: ParseResult,
        session: aiohttp.ClientSession,
        user_context: OptionalUserContext = None,
    ) -> AsyncIterator[eLabFTWRemoteEntryWrapper[RemoteDirectory]]:
        """
        List the root directory, i.e. "/".
        """
        # actually, the server does not need to be contacted to list entity types, but it makes sense to check
        # that it is alive and an actual eLabFTW instance is running to avoid giving the false impression that
        # things are working smoothly
        url = f"{endpoint.scheme}://{endpoint.netloc}/api/v2/info"
        async with session.get(
            url,
            allow_redirects=True,
            timeout=aiohttp.ClientTimeout(sock_connect=CONNECT_TIMEOUT, sock_read=READ_TIMEOUT),
        ) as response:
            try:
                is_valid = all(
                    (
                        response.status == 200,
                        content := await response.json(),
                        content.get("elabftw_version"),
                    )
                )
            except aiohttp.ContentTypeError:
                is_valid = False
            if not is_valid:
                raise ValidationError(err_msg="Invalid response from eLabFTW")

        experiments = eLabFTWRemoteEntryWrapper(
            RemoteDirectory(
                **{
                    "name": "Experiments",
                    "uri": f"{self.get_scheme()}://{self.get_prefix(user_context=user_context)}/experiments",
                    "path": "/experiments",
                    "class": "Directory",
                }
            )
        )
        resources = eLabFTWRemoteEntryWrapper(
            RemoteDirectory(
                **{
                    "name": "Resources",
                    "uri": f"{self.get_scheme()}://{self.get_prefix(user_context=user_context)}/resources",
                    "path": "/resources",
                    "class": "Directory",
                }
            )
        )

        yield experiments
        yield resources

    async def _yield_entities(
        self,
        entity_type: str,
        endpoint: ParseResult,
        session: aiohttp.ClientSession,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        order: Optional[str] = None,
        writable: bool = False,
        user_context: OptionalUserContext = None,
    ) -> AsyncIterator[eLabFTWRemoteEntryWrapper[RemoteDirectory]]:
        """List an entity type, i.e. either "/experiments" or "/resources"."""
        url = urljoin(
            f"{endpoint.scheme}://{endpoint.netloc}/",
            f"/api/v2/{entity_type.replace('resources', 'items')}",
        )

        class Params(TypedDict):
            order: str
            sort: str
            limit: int
            offset: int
            extended: NotRequired[str]
            q: NotRequired[str]

        params: Params = {
            "order": order or "id",
            "sort": "asc",
            "limit": min(MAX_ITEMS_PER_PAGE, limit) if limit is not None else MAX_ITEMS_PER_PAGE,
            "offset": offset or 0,
        }
        if writable:
            params.update({"extended": "locked:0"})
            # improvement: is there is a way to request only entities the user has permission to write to?
        if query:
            params.update({"q": query})

        content: List[dict] = [{}] * params["limit"]  # stores JSON responses (entities) from the server
        start, timeout = time(), False
        while len(content) >= params["limit"] and not (timeout := ((time() - start) >= PAGINATION_TIMEOUT)):
            entities: Dict[int, dict] = {}

            async with session.get(
                url,
                params={key: str(value) for key, value in params.items()},
                allow_redirects=True,
                timeout=aiohttp.ClientTimeout(sock_connect=CONNECT_TIMEOUT, sock_read=READ_TIMEOUT),
            ) as response:
                try:
                    status: int = response.status
                    content = await response.json()

                    def validate_and_register_entity(item, mapping: Dict[int, dict]) -> Literal[True]:
                        valid = isinstance(item, dict) and isinstance(item.get("id"), int)
                        if not valid:
                            raise ValidationError(err_msg="Invalid response from eLabFTW")
                        mapping[item["id"]] = item
                        return True

                    is_valid = all(
                        (
                            status == 200,
                            isinstance(content, list),
                            all(validate_and_register_entity(item, entities) for item in content),
                        )
                    )
                except aiohttp.ContentTypeError:
                    is_valid = False
                if not is_valid:
                    raise ValidationError(err_msg="Invalid response from eLabFTW")

            for entity in entities.values():
                yield eLabFTWRemoteEntryWrapper(
                    RemoteDirectory(
                        **{
                            "name": entity["title"],
                            "uri": (
                                f"{self.get_scheme()}://{self.get_prefix(user_context=user_context)}"
                                f"/{entity_type}/{entity['id']}"
                            ),
                            "path": f"/{entity_type}/{entity['id']}",
                            "class": "Directory",
                        }
                    ),
                    entity,
                )

            params["offset"] += params["limit"]

        if timeout:
            raise aiohttp.ServerTimeoutError

    async def _yield_attachments(
        self,
        entity_type: str,
        entity_id: str,
        endpoint: ParseResult,
        session: aiohttp.ClientSession,
        user_context: OptionalUserContext = None,
    ) -> AsyncIterator[eLabFTWRemoteEntryWrapper[RemoteFile]]:
        """List attachments of a specific entity, e.g. "/resources/48"."""
        url = urljoin(
            f"{endpoint.scheme}://{endpoint.netloc}/",
            f"/api/v2/{entity_type.replace('resources', 'items')}/{entity_id}",
        )
        try:
            async with session.get(
                url,
                allow_redirects=True,
                timeout=aiohttp.ClientTimeout(sock_connect=CONNECT_TIMEOUT, sock_read=READ_TIMEOUT),
            ) as response:
                try:
                    response_json = await response.json()
                    is_valid = True
                except aiohttp.ContentTypeError:
                    is_valid = False
                if not is_valid:
                    raise ValidationError(err_msg="Invalid response from eLabFTW")

            uploads = {upload["id"]: upload for upload in response_json.get("uploads", [])}
        except aiohttp.ClientResponseError as exception:
            if exception.status == 403:
                # cannot fetch items owned by someone else from the API but can do it from the browser, why?
                logging.exception(exception)
                uploads = {}
            else:
                raise exception

        for upload in uploads.values():
            yield eLabFTWRemoteEntryWrapper(
                RemoteFile(
                    **{
                        "name": upload["real_name"],
                        "uri": (
                            f"{self.get_scheme()}://{self.get_prefix(user_context=user_context)}"
                            f"/{entity_type}/{entity_id}/{upload['id']}"
                        ),
                        "path": f"/{entity_type}/{entity_id}/{upload['id']}",
                        "class": "File",
                        "size": upload["filesize"],
                        "ctime": datetime.fromisoformat(upload["created_at"]).astimezone(timezone.utc).isoformat(),
                    }
                ),
                upload,
            )

    def _write_from(
        self,
        target_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ) -> str:
        """
        Attach the file located at ``native_path`` on the filesystem to an eLabFTW resource or experiment with URI
        ``target_path``.

        :param target_path: URI of the experiment or resource that the file will be attached to, followed by the name it
            will be assigned; e.g.``elabftw://demo.elabftw.net/experiments/269/name``
        :type target_path: str
        :param native_path: The local file to upload, e.g. ``/tmp/myfile.txt``
        :type native_path: str
        :param user_context: A user context, defaults to ``None``
        :type user_context: OptionalUserContext
        :param opts: A set of options to exercise additional control over this method. Defaults to ``None``
        :type opts: Optional[FilesSourceOptions], optional
        :return: Path *assigned by eLabFTW* to the uploaded file.
        :rtype: str

        :raises requests.RequestException: When there is a connection error.
        :raises ValidationError: If the HTTP response from the eLabFTW server is invalid.
        :raises InvalidPath: After splitting `target_path` into the URI of the experiment or resource that the file will
                             be attached to and the name it will be assigned, the former is validated normally, and thus
                             this exception will be raised if the path constraints described in the docstring of
                             :class:`InvalidPath` are not satisfied. In addition, it will also be raised if the latter
                             is not a name but rather a relative path, meaning that `target_path` consists of more than
                             three components.
        :raises EntityExpected: When attempting to attach the file to the root "/" or an entity type.
        """
        session = self._create_session(options=opts, user_context=user_context)
        endpoint = self._get_endpoint(options=opts, user_context=user_context)

        target_path_obj = Path(target_path)
        attachment_name = target_path_obj.name
        try:
            entity_type, entity_id, attachment_id = split_path(str(target_path_obj.parent))
        finally:
            if len(target_path_obj.parts[1:]) > 3:
                raise InvalidPath(err_msg=InvalidPath.message_path_form % target_path_obj)
        if not all((entity_type, entity_id)):
            raise EntityExpected(err_msg="Expected an entity (an experiment or resource)")
        entity_type, entity_id = cast(str, entity_type), cast(str, entity_id)

        url = urljoin(
            f"{endpoint.scheme}://{endpoint.netloc}/",
            f"/api/v2/{entity_type.replace('resources', 'items')}/{entity_id}/uploads",
        )
        # cannot overwrite attachments by design, hence disabled
        # if attachment_id:
        #     url += f"/{attachment_id}"

        with open(native_path, "rb") as file:
            response = session.post(
                url,
                files={"file": (attachment_name, file)},
                timeout=(CONNECT_TIMEOUT, READ_TIMEOUT),
            )
            response.raise_for_status()

            try:
                location = urlparse(response.headers["location"])
                is_valid = all(
                    (
                        response.status_code == 201,
                        location.netloc == endpoint.netloc,
                        (match := (re.match(r"^/api/v2/(experiments|items)/([0-9]+)/uploads/([0-9]+)$", location.path)))
                        is not None,
                    )
                )
            except KeyError:
                is_valid = False
            if not is_valid:
                raise ValidationError(err_msg="Invalid response from eLabFTW")
            match = cast(re.Match, match)

        entity_type, entity_id, attachment_id = match.groups()
        entity_type = entity_type.replace("items", "resources")

        return f"/{entity_type}/{entity_id}/{attachment_id}"

    def _realize_to(
        self,
        source_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        """
        Save the file attachment from an eLabFTW resource or experiment located at ``source_path`` to ``native_path``.

        :param source_path: URI of the file ``elabftw://demo.elabftw.net/experiments/269/69`` to download from eLabFTW
        :type source_path: str
        :param native_path: The path on the filesystem to save the file to, e.g. ``/tmp/myfile.txt``
        :type native_path: str
        :param user_context: A user context, defaults to ``None``
        :type user_context: OptionalUserContext
        :param opts: A set of options to exercise additional control over this method. Defaults to ``None``

        :raises requests.RequestException: When there is a connection error.
        :raises ValidationError: If the HTTP response from the eLabFTW server is invalid.
        :raises AttachmentExpected: When referencing an entity type, an entity or the root rather than an attachment.
        """
        session = self._create_session(options=opts, user_context=user_context)
        endpoint = self._get_endpoint(options=opts, user_context=user_context)

        entity_type, entity_id, attachment_id = split_path(source_path)
        if not all((entity_type, entity_id, attachment_id)):
            raise AttachmentExpected(err_msg="Expected a file attached to an experiment or resource")
        entity_type, entity_id, attachment_id = cast(str, entity_type), cast(str, entity_id), cast(str, attachment_id)

        url = urljoin(
            f"{endpoint.scheme}://{endpoint.netloc}/",
            f"/api/v2/{entity_type.replace('resources', 'items')}/{entity_id}/uploads/{attachment_id}"
            f"?format=binary",
        )
        try:
            with (
                session.get(
                    url,
                    timeout=(CONNECT_TIMEOUT, READ_TIMEOUT),
                    stream=True,
                ) as response,
                open(native_path, "wb") as file,
            ):
                response.raise_for_status()
                for chunk in response.iter_content(512):
                    file.write(chunk)
        except Exception as exception:
            Path(native_path).unlink(missing_ok=True)
            raise exception


def split_path(path: str) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """
    Split and validate an eLabFTW path.

    Split an eLabFTW path into its parts, and ensure that it satisfies the constraints this module imposes on it (see
    docstring of :class:`InvalidPath`).

    :param path: A path representing an entity type, an entity, or a file attachment in eLabFTW.
    :type path: str

    :returns: The path passed as an argument split into three parts: ``entity_type``, ``entity_id``, and
              ``attachment_id``. ``None`` is returned in place of missing parts.
    :rtype: Tuple[Optional[str], Optional[str], Optional[str]]

    :raises InvalidPath: Path constraints described in the docstring of :class:`InvalidPath` are not satisfied.
    """
    path_obj = Path(path)

    if not path_obj.is_absolute():
        raise InvalidPath(err_msg=InvalidPath.message_path_absolute % path_obj)

    parts = path_obj.parts[1:]
    if len(parts) > 3:
        raise InvalidPath(err_msg=InvalidPath.message_path_form % path_obj)
    entity_type, entity_id, attachment_id = (
        # right pad `parts` with three `None`s
        tuple(parts)
        + (None,) * (3 - len(parts))
    )

    if entity_type not in (None, "experiments", "resources"):
        raise InvalidPath(err_msg=InvalidPath.message_path_entity_type % path_obj)

    if entity_id:
        try:
            if not int(entity_id) >= 0:
                raise ValueError
        except ValueError as exception:
            raise InvalidPath(err_msg=InvalidPath.message_path_entity_id % path_obj) from exception

    if attachment_id:
        try:
            if not int(attachment_id) >= 0:
                raise ValueError
        except ValueError as exception:
            raise InvalidPath(err_msg=InvalidPath.message_path_attachment_id % path_obj) from exception

    return entity_type, entity_id, attachment_id


class eLabFTWFilesSourceException(  # noqa
    ABC,
    Exception,
):
    """Base class for exceptions raised when `eLabFTWFilesSource` encounters a problem."""


class InvalidPath(
    galaxy_exceptions.MessageException,
    eLabFTWFilesSourceException,
):
    """
    Raised when an invalid path is provided.

    Valid paths are of the form `/entity_type/entity_id/attachment_id`, where:
    - `entity_type` is either 'experiments' or 'resources'
    - `entity_id` is the id (an integer) of an experiment or resource
    - `attachment_id` is the id (an integer) of an attachment
    """

    message_path_form = (
        # fmt: off
        "path '%' is invalid, paths must be of the form "
        "`/entity_type/entity_id/attachment_id`, where:"
        + dedent("""
            - `entity_type` is either 'experiments' or 'resources'
            - `entity_id` is the id of an experiment or resource
            - `attachment_id` is the id of an attachment
        """[1:])
        # fmt: on
    )
    message_path_absolute = "path '%' is invalid, paths must be absolute"
    message_path_entity_type = "path '%' is invalid, paths must start with /experiments or /resources"
    message_path_entity_id = (
        "path '%' is invalid, the entity id (second part of the path) must be a non-negative integer"
    )
    message_path_attachment_id = (
        "path '%' is invalid, the attachment id (third part of the path) must be a non-negative integer"
    )


class ResourceNotFound(
    galaxy_exceptions.ObjectNotFound,
    eLabFTWFilesSourceException,
):
    """
    Raised when attempting to access a non-existing experiment, resource or attachment.
    """


class DirectoryExpected(
    galaxy_exceptions.MessageException,
    eLabFTWFilesSourceException,
    ValueError,
):
    """
    Raised when referencing a file attachment where a path referencing an entity type, entity or the root is required.

    For example, it would be raised when attempting to list the contents of a file attachment. Only the root, entity
    types and entities themselves have "contents" (entity types, entities and file attachments respectively).
    """


class EntityExpected(DirectoryExpected):
    """
    Raised when referencing the root, an entity type or an attachment where a path referencing an entity is expected.

    For example, attempting to save a file to an entity type using `_write_from()` raises this exception.
    """


class AttachmentExpected(
    galaxy_exceptions.MessageException,
    eLabFTWFilesSourceException,
    ValueError,
):
    """
    Raised when referencing an entity type, an entity or the root where a path referencing an attachment is required.
    """


class ValidationError(
    galaxy_exceptions.MessageException,
    eLabFTWFilesSourceException,
):
    """
    Raised when validation of a response from eLabFTW fails.

    API responses from eLabFTW are considered untrusted and thus validated before the file source makes use of them.
    This exception will be raised when an invalid response is detected.
    """



import logging
import os
from collections import defaultdict
from typing import (
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Protocol,
    Set,
)

from galaxy import exceptions
from galaxy.files.sources import (
    BaseFilesSource,
    FilesSourceProperties,
    PluginKind,
)
from galaxy.util.dictifiable import Dictifiable
from galaxy.util.plugin_config import (
    plugin_source_from_dict,
    plugin_source_from_path,
    PluginConfigSource,
    PluginConfigsT,
)
from .plugins import (
    FileSourcePluginLoader,
    FileSourcePluginsConfig,
)

log = logging.getLogger(__name__)


class FileSourcePath(NamedTuple):
    file_source: BaseFilesSource
    path: str


class FileSourceScore(NamedTuple):
    file_source: BaseFilesSource
    score: int


class NoMatchingFileSource(Exception):
    pass


class UserDefinedFileSources(Protocol):
    """Entry-point for Galaxy to inject user-defined object stores.

    Supplied object of this class is used to write out concrete
    description of file sources when serializing all file sources
    available to a user.
    """

    def validate_uri_root(self, uri: str, user_context: "FileSourcesUserContext") -> None:
        pass

    def find_best_match(self, url: str) -> Optional[FileSourceScore]:
        pass

    def user_file_sources_to_dicts(
        self,
        for_serialization: bool,
        user_context: "FileSourcesUserContext",
        browsable_only: Optional[bool] = False,
        include_kind: Optional[Set[PluginKind]] = None,
        exclude_kind: Optional[Set[PluginKind]] = None,
    ) -> List[FilesSourceProperties]:
        """Write out user file sources as list of config dictionaries."""
        # config_dicts: List[FilesSourceProperties] = []
        # for file_source in self.user_file_sources():
        #     as_dict = file_source.to_dict(for_serialization=for_serialization, user_context=user_context)
        #     config_dicts.append(as_dict)
        # return config_dicts


class NullUserDefinedFileSources(UserDefinedFileSources):

    def validate_uri_root(self, uri: str, user_context: "FileSourcesUserContext") -> None:
        return None

    def find_best_match(self, url: str) -> Optional[FileSourceScore]:
        return None

    def user_file_sources_to_dicts(
        self,
        for_serialization: bool,
        user_context: "FileSourcesUserContext",
        browsable_only: Optional[bool] = False,
        include_kind: Optional[Set[PluginKind]] = None,
        exclude_kind: Optional[Set[PluginKind]] = None,
    ) -> List[FilesSourceProperties]:
        return []


def _ensure_user_defined_file_sources(
    user_defined_file_sources: Optional[UserDefinedFileSources] = None,
) -> UserDefinedFileSources:
    if user_defined_file_sources is not None:
        return user_defined_file_sources
    else:
        return NullUserDefinedFileSources()


class ConfiguredFileSourcesConf:
    conf_dict: Optional[PluginConfigsT]
    conf_file: Optional[str]

    def __init__(self, conf_dict: Optional[PluginConfigsT] = None, conf_file: Optional[str] = None):
        self.conf_dict = conf_dict
        self.conf_file = conf_file

    @staticmethod
    def from_app_config(config):
        config_file = config.file_sources_config_file
        config_dict = None
        if not config_file or not os.path.exists(config_file):
            config_file = None
            config_dict = config.file_sources
        return ConfiguredFileSourcesConf(config_dict, config_file)


class ConfiguredFileSources:
    """Load plugins and resolve Galaxy URIs to FileSource objects."""

    _file_sources: List[BaseFilesSource]
    _plugin_loader: FileSourcePluginLoader
    _user_defined_file_sources: UserDefinedFileSources

    def __init__(
        self,
        file_sources_config: FileSourcePluginsConfig,
        configured_file_source_conf: Optional[ConfiguredFileSourcesConf] = None,
        load_stock_plugins: bool = False,
        plugin_loader: Optional[FileSourcePluginLoader] = None,
        user_defined_file_sources: Optional[UserDefinedFileSources] = None,
    ):
        self._file_sources_config = file_sources_config
        self._plugin_loader = plugin_loader or FileSourcePluginLoader()
        self._user_defined_file_sources = _ensure_user_defined_file_sources(user_defined_file_sources)
        file_sources: List[BaseFilesSource] = []
        if configured_file_source_conf is None:
            configured_file_source_conf = ConfiguredFileSourcesConf(conf_dict=[])
        if configured_file_source_conf.conf_file is not None:
            file_sources = self._load_plugins_from_file(configured_file_source_conf.conf_file)
        elif configured_file_source_conf.conf_dict is not None:
            plugin_source = plugin_source_from_dict(configured_file_source_conf.conf_dict)
            file_sources = self._parse_plugin_source(plugin_source)
        else:
            file_sources = []
        custom_sources_configured = len(file_sources) > 0 or (user_defined_file_sources is not None)
        if load_stock_plugins:
            stock_file_source_conf_dict = []

            def _ensure_loaded(plugin_type):
                for file_source in file_sources:
                    if file_source.plugin_type == plugin_type:
                        return
                stock_file_source_conf_dict.append({"type": plugin_type})

            _ensure_loaded("http")
            _ensure_loaded("base64")
            _ensure_loaded("drs")
            _ensure_loaded("remoteZip")

            if file_sources_config.ftp_upload_dir is not None:
                _ensure_loaded("gxftp")
            if file_sources_config.library_import_dir is not None:
                _ensure_loaded("gximport")
            if file_sources_config.user_library_import_dir is not None:
                _ensure_loaded("gxuserimport")

            if stock_file_source_conf_dict:
                stock_plugin_source = plugin_source_from_dict(stock_file_source_conf_dict)
                # insert at beginning instead of append so FTP and library import appear
                # at the top of the list (presumably the most common options). Admins can insert
                # these explicitly for greater control.
                file_sources = self._parse_plugin_source(stock_plugin_source) + file_sources

        self._file_sources = file_sources
        self.custom_sources_configured = custom_sources_configured

    def _load_plugins_from_file(self, conf_file: str):
        plugin_source = plugin_source_from_path(conf_file)
        return self._parse_plugin_source(plugin_source)

    def _parse_plugin_source(self, plugin_source: PluginConfigSource):
        return self._plugin_loader.load_plugins(plugin_source, self._file_sources_config)

    def find_best_match(self, url: str) -> Optional[BaseFilesSource]:
        """Returns the best matching file source for handling a particular url. Each filesource scores its own
        ability to match a particular url, and the highest scorer with a score > 0 is selected."""
        scores = [FileSourceScore(file_source, file_source.score_url_match(url)) for file_source in self._file_sources]
        user_best_score = self._user_defined_file_sources.find_best_match(url)
        if user_best_score is not None:
            scores.append(user_best_score)
        scores.sort(key=lambda f: f.score, reverse=True)
        return next((fsscore.file_source for fsscore in scores if fsscore.score > 0), None)

    def get_file_source_path(self, uri):
        """Parse uri into a FileSource object and a path relative to its base."""
        if "://" not in uri:
            raise exceptions.RequestParameterInvalidException(f"Invalid uri [{uri}]")
        file_source = self.find_best_match(uri)
        if not file_source:
            raise exceptions.RequestParameterInvalidException(f"Could not find handler for URI [{uri}]")
        path = file_source.to_relative_path(uri)
        return FileSourcePath(file_source, path)

    def validate_uri_root(self, uri: str, user_context: "FileSourcesUserContext"):
        # validate a URI against Galaxy's configuration, environment, and the current
        # user. Throw appropriate exception if there is a problem with the files source
        # referenced by the URI.
        if uri.startswith("gxuserimport://"):
            user_login = user_context.email
            user_base_dir = self._file_sources_config.user_library_import_dir
            if user_base_dir is None:
                raise exceptions.ConfigDoesNotAllowException(
                    "The configuration of this Galaxy instance does not allow upload from user directories."
                )
            if user_login is None:
                raise exceptions.AuthenticationRequired("Must be logged in to use this feature.")
            full_import_dir = os.path.join(user_base_dir, user_login)
            if not os.path.exists(full_import_dir):
                raise exceptions.ObjectNotFound("Your user import directory does not exist.")
        elif uri.startswith("gximport://"):
            base_dir = self._file_sources_config.library_import_dir
            if base_dir is None:
                raise exceptions.ConfigDoesNotAllowException(
                    "The configuration of this Galaxy instance does not allow usage of import directory."
                )
        elif uri.startswith("gxftp://"):
            user_ftp_base_dir = self._file_sources_config.ftp_upload_dir
            if user_ftp_base_dir is None:
                raise exceptions.ConfigDoesNotAllowException(
                    "The configuration of this Galaxy instance does not allow upload from FTP directories."
                )
            user_ftp_dir = user_context.ftp_dir
            if not user_ftp_dir or not os.path.exists(user_ftp_dir):
                raise exceptions.ObjectNotFound(
                    "Your FTP directory does not exist, attempting to upload files to it may cause it to be created."
                )
        self._user_defined_file_sources.validate_uri_root(uri, user_context)

    def looks_like_uri(self, path_or_uri):
        # is this string a URI this object understands how to realize
        file_source = self.find_best_match(path_or_uri)
        if file_source:
            return True
        else:
            return False

    def plugins_to_dict(
        self,
        for_serialization: bool = False,
        user_context: "OptionalUserContext" = None,
        browsable_only: Optional[bool] = False,
        include_kind: Optional[Set[PluginKind]] = None,
        exclude_kind: Optional[Set[PluginKind]] = None,
    ) -> List[FilesSourceProperties]:
        rval: List[FilesSourceProperties] = []
        for file_source in self._file_sources:
            if not file_source.user_has_access(user_context):
                continue
            if include_kind and file_source.plugin_kind not in include_kind:
                continue
            if exclude_kind and file_source.plugin_kind in exclude_kind:
                continue
            if browsable_only and not file_source.get_browsable():
                continue
            el = file_source.to_dict(for_serialization=for_serialization, user_context=user_context)
            rval.append(el)
        if user_context:
            rval.extend(
                self._user_defined_file_sources.user_file_sources_to_dicts(
                    for_serialization,
                    user_context,
                    browsable_only=browsable_only,
                    include_kind=include_kind,
                    exclude_kind=exclude_kind,
                )
            )
        return rval

    def to_dict(self, for_serialization: bool = False, user_context: "OptionalUserContext" = None) -> Dict[str, Any]:
        return {
            "file_sources": self.plugins_to_dict(for_serialization=for_serialization, user_context=user_context),
            "config": self._file_sources_config.to_dict(),
        }

    @staticmethod
    def from_dict(as_dict, load_stock_plugins=False):
        if as_dict is not None:
            sources_as_dict = as_dict["file_sources"]
            config_as_dict = as_dict["config"]
            file_sources_config = FileSourcePluginsConfig.from_dict(config_as_dict)
        else:
            sources_as_dict = []
            file_sources_config = FileSourcePluginsConfig()
        configured_file_sources_conf = ConfiguredFileSourcesConf(conf_dict=sources_as_dict)
        return ConfiguredFileSources(
            file_sources_config, configured_file_sources_conf, load_stock_plugins=load_stock_plugins
        )


class NullConfiguredFileSources(ConfiguredFileSources):
    def __init__(
        self,
    ):
        super().__init__(FileSourcePluginsConfig(), ConfiguredFileSourcesConf(conf_dict=[]))


class DictifiableFilesSourceContext(Protocol):
    @property
    def role_names(self) -> Set[str]: ...

    @property
    def group_names(self) -> Set[str]: ...

    @property
    def file_sources(self) -> ConfiguredFileSources: ...

    def to_dict(
        self, view: str = "collection", value_mapper: Optional[Dict[str, Callable]] = None
    ) -> Dict[str, Any]: ...


class FileSourceDictifiable(Dictifiable, DictifiableFilesSourceContext):
    dict_collection_visible_keys = ("email", "username", "ftp_dir", "preferences", "is_admin")

    def to_dict(self, view="collection", value_mapper: Optional[Dict[str, Callable]] = None) -> Dict[str, Any]:
        rval = super().to_dict(view=view, value_mapper=value_mapper)
        rval["role_names"] = list(self.role_names)
        rval["group_names"] = list(self.group_names)
        return rval


class FileSourcesUserContext(DictifiableFilesSourceContext, Protocol):

    @property
    def email(self) -> Optional[str]: ...

    @property
    def username(self) -> Optional[str]: ...

    @property
    def ftp_dir(self) -> Optional[str]: ...

    @property
    def preferences(self) -> Dict[str, Any]: ...

    @property
    def is_admin(self) -> bool: ...

    @property
    def user_vault(self) -> Dict[str, Any]: ...

    @property
    def app_vault(self) -> Dict[str, Any]: ...

    @property
    def anonymous(self) -> bool: ...


OptionalUserContext = Optional[FileSourcesUserContext]


class ProvidesFileSourcesUserContext(FileSourcesUserContext, FileSourceDictifiable):
    """Implement a FileSourcesUserContext from a Galaxy ProvidesUserContext (e.g. trans)."""

    def __init__(self, trans, **kwargs):
        self.trans = trans

    @property
    def email(self) -> Optional[str]:
        user = self.trans.user
        return user and user.email

    @property
    def username(self) -> Optional[str]:
        user = self.trans.user
        return user and user.username

    @property
    def ftp_dir(self):
        return self.trans.user_ftp_dir

    @property
    def preferences(self):
        user = self.trans.user
        return user and user.extra_preferences or defaultdict(lambda: None)

    @property
    def role_names(self) -> Set[str]:
        """The set of role names of this user."""
        user = self.trans.user
        role_names = set()
        if user:
            role_names = {ura.role.name for ura in user.roles}
            role_names.add(user.email)  # User's private role may have a generic name, so add user's email explicitly.
        return role_names

    @property
    def group_names(self) -> Set[str]:
        """The set of group names to which this user belongs."""
        user = self.trans.user
        return {ugr.group.name for ugr in user.groups} if user else set()

    @property
    def is_admin(self):
        """Whether this user is an administrator."""
        return self.trans.user_is_admin

    @property
    def user_vault(self):
        """User vault namespace"""
        user_vault = self.trans.user_vault
        return user_vault or defaultdict(lambda: None)

    @property
    def app_vault(self):
        """App vault namespace"""
        vault = self.trans.app.vault
        return vault or defaultdict(lambda: None)

    @property
    def file_sources(self):
        return self.trans.app.file_sources

    @property
    def anonymous(self) -> bool:
        return self.trans.anonymous


class DictFileSourcesUserContext(FileSourcesUserContext, FileSourceDictifiable):
    def __init__(self, **kwd):
        self._kwd = kwd

    @property
    def email(self):
        return self._kwd.get("email")

    @property
    def username(self) -> Optional[str]:
        return self._kwd.get("username")

    @property
    def ftp_dir(self):
        return self._kwd.get("user_ftp_dir")

    @property
    def preferences(self):
        return self._kwd.get("preferences")

    @property
    def role_names(self):
        return set(self._kwd.get("role_names", []))

    @property
    def group_names(self):
        return set(self._kwd.get("group_names", []))

    @property
    def is_admin(self):
        return self._kwd.get("is_admin")

    @property
    def user_vault(self):
        return self._kwd.get("user_vault")

    @property
    def app_vault(self):
        return self._kwd.get("app_vault")

    @property
    def file_sources(self):
        return self._kwd.get("file_sources")

    @property
    def anonymous(self) -> bool:
        return not bool(self._kwd.get("username"))



import functools
import logging
import os
from typing import (
    cast,
    List,
    Optional,
    Tuple,
)

from typing_extensions import (
    NotRequired,
    Unpack,
)

from galaxy import exceptions
from galaxy.files import OptionalUserContext
from . import (
    AnyRemoteEntry,
    FilesSourceOptions,
    FilesSourceProperties,
)

try:
    import s3fs
except ImportError:
    s3fs = None

from . import BaseFilesSource

DEFAULT_ENFORCE_SYMLINK_SECURITY = True
DEFAULT_DELETE_ON_REALIZE = False

log = logging.getLogger(__name__)


class S3FsFilesSourceProperties(FilesSourceProperties, total=False):
    bucket: str
    endpoint_url: int
    user: str
    passwd: str
    listings_expiry_time: NotRequired[Optional[int]]
    client_kwargs: dict  # internally computed. Should not be specified in config file


class S3FsFilesSource(BaseFilesSource):
    plugin_type = "s3fs"

    def __init__(self, **kwd: Unpack[S3FsFilesSourceProperties]):
        if s3fs is None:
            raise Exception("Package s3fs unavailable but required for this file source plugin.")
        props: S3FsFilesSourceProperties = cast(S3FsFilesSourceProperties, self._parse_common_config_opts(kwd))
        file_sources_config = self._file_sources_config
        if (
            props.get("listings_expiry_time") is None
            and file_sources_config
            and file_sources_config.listings_expiry_time
        ):
            if file_sources_config.listings_expiry_time:
                props["listings_expiry_time"] = file_sources_config.listings_expiry_time
        # There is a possibility that the bucket name could be parameterized: e.g.
        # bucket: ${user.preferences['generic_s3|bucket']}
        # that's ok, because we evaluate the bucket name again later. The bucket property here will only
        # be used by `score_url_match`. In the case of a parameterized bucket name, we will always return
        # a score of 4 as the url will only match the s3:// part.
        self._bucket = props.get("bucket", "")
        self._endpoint_url = props.pop("endpoint_url", None)
        self._props = props
        if self._endpoint_url:
            self._props.update({"client_kwargs": {"endpoint_url": self._endpoint_url}})

    def _list(
        self,
        path="/",
        recursive=True,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        sort_by: Optional[str] = None,
    ) -> Tuple[List[AnyRemoteEntry], int]:
        _props = self._serialization_props(user_context)
        # we need to pop the 'bucket' here, because the argument is not recognised in a downstream function
        _bucket_name = _props.pop("bucket", "")
        fs = self._open_fs(props=_props, opts=opts)
        if recursive:
            res: List[AnyRemoteEntry] = []
            bucket_path = self._bucket_path(_bucket_name, path)
            for p, dirs, files in fs.walk(bucket_path, detail=True):
                to_dict = functools.partial(self._resource_info_to_dict, p)
                res.extend(map(to_dict, dirs.values()))
                res.extend(map(to_dict, files.values()))
            return res, len(res)
        else:
            bucket_path = self._bucket_path(_bucket_name, path)
            try:
                res = fs.ls(bucket_path, detail=True)
            except Exception as e:
                raise exceptions.MessageException(f"Error listing {bucket_path}: {e}")
            to_dict = functools.partial(self._resource_info_to_dict, path)
            return list(map(to_dict, res)), len(res)

    def _realize_to(
        self,
        source_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        _props = self._serialization_props(user_context)
        _bucket_name = _props.pop("bucket", "")
        fs = self._open_fs(props=_props, opts=opts)
        bucket_path = self._bucket_path(_bucket_name, source_path)
        fs.download(bucket_path, native_path)

    def _write_from(
        self,
        target_path,
        native_path,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        _props = self._serialization_props(user_context)
        _bucket_name = _props.pop("bucket", "")
        fs = self._open_fs(props=_props, opts=opts)
        bucket_path = self._bucket_path(_bucket_name, target_path)
        fs.upload(native_path, bucket_path)

    def _bucket_path(self, bucket_name: str, path: str):
        if path.startswith("s3://"):
            return path.replace("s3://", "")
        elif not path.startswith("/"):
            path = f"/{path}"
        return f"{bucket_name}{path}"

    def _open_fs(self, props: FilesSourceProperties, opts: Optional[FilesSourceOptions] = None):
        extra_props = opts.extra_props or {} if opts else {}
        fs = s3fs.S3FileSystem(**{**props, **extra_props})
        return fs

    def _resource_info_to_dict(self, dir_path: str, resource_info) -> AnyRemoteEntry:
        name = os.path.basename(resource_info["name"])
        path = os.path.join(dir_path, name)
        uri = self.uri_from_path(path)
        if resource_info["type"] == "directory":
            return {"class": "Directory", "name": name, "uri": uri, "path": path}
        else:
            return {
                "class": "File",
                "name": name,
                "size": resource_info["size"],
                # should this be mtime...
                "ctime": self.to_dict_time(resource_info["LastModified"]),
                "uri": uri,
                "path": path,
            }

    def _serialization_props(self, user_context: OptionalUserContext = None) -> S3FsFilesSourceProperties:
        effective_props = {}
        for key, val in self._props.items():
            effective_props[key] = self._evaluate_prop(val, user_context=user_context)
        return cast(S3FsFilesSourceProperties, effective_props)

    def score_url_match(self, url: str):
        # For security, we need to ensure that a partial match doesn't work. e.g. s3://{bucket}something/myfiles
        if self._bucket and (url.startswith(f"s3://{self._bucket}/") or url == f"s3://{self._bucket}"):
            return len(f"s3://{self._bucket}")
        elif not self._bucket and url.startswith("s3://"):
            return len("s3://")
        else:
            return super().score_url_match(url)


__all__ = ("S3FsFilesSource",)



import base64
import logging
from typing import Optional

from typing_extensions import Unpack

from galaxy.files import OptionalUserContext
from . import (
    BaseFilesSource,
    FilesSourceOptions,
    FilesSourceProperties,
    PluginKind,
)

log = logging.getLogger(__name__)


class Base64FilesSource(BaseFilesSource):
    plugin_type = "base64"
    plugin_kind = PluginKind.stock

    def __init__(self, **kwd: Unpack[FilesSourceProperties]):
        kwds: FilesSourceProperties = dict(
            id="_base64",
            label="Base64 encoded string",
            doc="Base64 string handler",
            writable=False,
        )
        kwds.update(kwd)
        props = self._parse_common_config_opts(kwds)
        self._props = props

    def _realize_to(
        self,
        source_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        with open(native_path, "wb") as temp:
            temp.write(base64.b64decode(source_path[len("base64://") :]))
            temp.flush()

    def _write_from(
        self,
        target_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        raise NotImplementedError()

    def score_url_match(self, url: str):
        if url.startswith("base64://"):
            return len("base64://")
        else:
            return 0

    def _serialization_props(self, user_context: OptionalUserContext = None):
        effective_props = {}
        for key, val in self._props.items():
            effective_props[key] = self._evaluate_prop(val, user_context=user_context)
        return effective_props


__all__ = ("Base64FilesSource",)



try:
    from fs.dropboxfs import DropboxFS
except ImportError:
    DropboxFS = None

from typing import (
    Optional,
    Union,
)

from galaxy.exceptions import (
    AuthenticationRequired,
    MessageException,
)
from . import (
    FilesSourceOptions,
    FilesSourceProperties,
)
from ._pyfilesystem2 import PyFilesystem2FilesSource


class DropboxFilesSource(PyFilesystem2FilesSource):
    plugin_type = "dropbox"
    required_module = DropboxFS
    required_package = "fs.dropboxfs"

    def _open_fs(self, user_context=None, opts: Optional[FilesSourceOptions] = None):
        props = self._serialization_props(user_context)
        extra_props: Union[FilesSourceProperties, dict] = opts.extra_props or {} if opts else {}
        # accessToken has been renamed to access_token in fs.dropboxfs 1.0
        if "accessToken" in props:
            props["access_token"] = props.pop("accessToken")
        if "oauth2_access_token" in props:
            props["access_token"] = props.pop("oauth2_access_token")

        try:
            handle = DropboxFS(**{**props, **extra_props})
            return handle
        except Exception as e:
            # This plugin might raise dropbox.dropbox_client.BadInputException
            # which is not a subclass of fs.errors.FSError
            if "OAuth2" in str(e):
                raise AuthenticationRequired(
                    f"Permission Denied. Reason: {e}. Please check your credentials in your preferences for {self.label}."
                )
            raise MessageException(f"Error connecting to Dropbox. Reason: {e}")


__all__ = ("DropboxFilesSource",)



"""
Galaxy FilesSource implementation for RSpace.

This module implements a FilesSource that interacts with an RSpace [1] instance. RSpace is an electronic lab notebook
(ELN) designed for documenting, managing and sharing scientific data. The central concepts in RSpace are the Workspace,
the Inventory and the Gallery.

The Workspace's role is organizing and managing research activities. It allows creating documents, folders and research
notebooks [2]. The Inventory is a tool to keep track of laboratory resources, such as reagents, samples and equipment
[3]. The Gallery is where any research data represented as files is stored [4].

The FilesSource implemented by this module accesses the Gallery of an RSpace instance. The Gallery is organized in
categories (e.g. Images, Audio, Documents, Chemistry), where each category is itself a tree structure with folders and
files just like a regular filesystem, although there is a key difference: categories and their subfolders can only
contain files of certain types (e.g. jpeg, bmp, png, etc. for the Images category).

This implementation is based on the ``rspace-client-python`` library [5], which provides a PyFilesystem2 [6] interface
to interact with RSpace's Gallery (available as ``GalleryFilesystem`` in the ``rspace_client.eln.fs`` module). It
consists just of a minimal wrapper (based on the base ``PyFilesystem2FilesSource`` provided by Galaxy) around the
``GalleryFilesystem`` class.

References:

- [1] https://www.researchspace.com/
- [2] https://documentation.researchspace.com/article/bzgr8ea9e3-the-workspace
- [3] https://documentation.researchspace.com/category/zpizk20kgx-inventory
- [4] https://documentation.researchspace.com/article/sl6mo1i9do-the-gallery
- [5] https://github.com/rspace-os/rspace-client-python
- [6] https://docs.pyfilesystem.org/
"""

import datetime
import os.path
from types import MethodType
from typing import (
    Any,
    BinaryIO,
    cast,
    IO,
    Optional,
    Union,
)

from galaxy.files import OptionalUserContext
from . import (
    AnyRemoteEntry,
    FilesSourceOptions,
    FilesSourceProperties,
    RemoteDirectory,
    RemoteFile,
)
from ._pyfilesystem2 import PyFilesystem2FilesSource

try:
    from rspace_client.eln.fs import GalleryFilesystem as RSpaceGalleryFilesystem
except ImportError:
    RSpaceGalleryFilesystem = None

__all__ = ("RSpaceFilesSource",)


class FakedNameIO:
    """
    A wrapper around a file-like object that fakes the name attribute.

    The name attribute is used by `requests` (that `rspace-client-python` depends on) to determine the mime type. RSpace
    will reject uploads to the Gallery that do not match the file types accepted by the target folder (e.g. no `.dat`
    files allowed in the "Images" folder).

    Galaxy datasets are stored in the filesystem with the `.dat` extension. Therefore, the ability to fake the name (and
    thus the extension) is essential to be able to upload files to any RSpace folder other than "Miscellaneous" without
    having to alter the `rspace-client-python` library itself.
    """

    def __init__(self, handle: IO, name: Optional[str] = None):
        """Initialize the wrapper from an existing file-like object."""
        self._handle = handle
        self._name = name

    @property
    def name(self) -> str:
        """Retrieve the faked name if it was provided, otherwise retrieve the original file name."""
        return self._name or self._handle.name

    def __getattr__(self, name):
        """Retrieve any other attribute."""
        return getattr(self._handle, name)


if RSpaceGalleryFilesystem is not None:

    class PatchedRSpaceGalleryFilesystem(RSpaceGalleryFilesystem):
        """
        Patch RSpaceGalleryFilesystem to keep a record of the RSpace global id of the most recently uploaded file.
        """

        upload_global_id: str
        # The RSpace global id of the most recently uploaded file.

        _upload_response: dict
        # The response from the RSpace's API (in JSON format) after uploading a file.

        def __init__(self, *args, **kwargs):
            """
            Initialize the RSpaceGalleryFilesystem and patch the `upload_file()` method to save the upload response.
            """
            super().__init__(*args, **kwargs)

            # patch the `upload_file()` method to save the upload response
            eln_client_upload_file_method = self.eln_client.upload_file

            def upload_file(self_, file, folder_id=None, caption=None):
                self._upload_response = eln_client_upload_file_method(file, folder_id=folder_id, caption=caption)

            self.eln_client.upload_file = MethodType(upload_file, self.eln_client)

        def upload(self, path: str, file: BinaryIO, chunk_size: Optional[int] = None, **options: Any) -> None:
            """
            Patch the `upload()` method to retrieve the global id from the saved upload response.
            """
            super().upload(path, file, chunk_size, **options)
            self.upload_global_id = self._upload_response["globalId"]


class RSpaceFilesSource(PyFilesystem2FilesSource):
    plugin_type = "rspace"
    required_module = RSpaceGalleryFilesystem
    required_package = "rspace_client.eln.fs"

    _upload_global_id: str
    # The RSpace global id of the most recently uploaded file.

    def _open_fs(self, user_context=None, opts: Optional[FilesSourceOptions] = None) -> RSpaceGalleryFilesystem:
        """
        Instantiate a PyFilesystem2 FS object for RSpace's Gallery.
        """
        props = self._serialization_props(user_context)
        extra_props: Union[FilesSourceProperties, dict] = opts.extra_props or {} if opts else {}
        all_props: dict = {**props, **extra_props}

        endpoint = all_props["endpoint"]
        api_key = all_props["api_key"]

        # patch the `upload()` method to keep track of the global id of the most recently uploaded file, change the
        # target path and fake the name of the file
        gallery_fs = PatchedRSpaceGalleryFilesystem(endpoint, api_key)
        gallery_fs_upload_method = gallery_fs.upload

        def upload(self_, path: str, file: BinaryIO, chunk_size: Optional[int] = None, **options: Any) -> None:
            gallery_fs_upload_method(
                os.path.dirname(path),
                cast(BinaryIO, FakedNameIO(file, name=os.path.basename(path))),
                chunk_size,
                **options,
            )
            self._upload_global_id = gallery_fs.upload_global_id

        gallery_fs.upload = MethodType(upload, gallery_fs)  # type: ignore[method-assign]

        return gallery_fs

    def _write_from(
        self,
        target_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ) -> str:
        """
        Save a file to the RSpace Gallery.
        """
        target_directory = os.path.dirname(target_path)
        super()._write_from(target_path, native_path, user_context=user_context, opts=opts)
        return os.path.join(target_directory, self._upload_global_id)

    def _resource_info_to_dict(self, dir_path, resource_info) -> AnyRemoteEntry:
        """
        Convert PyFilesystem2 resource information objects to dictionaries representing directories and files.

        Display names for categories, directories and files in the RSpace Gallery are made available within the custom
        namespace "rspace" (as well as creation dates). Names are available in the "basic" namespace are the internal
        RSpace names, which are unique. Thus, those names from the "basic" namespace are used to generate the URIs for
        the directories and files (remote entries).
        """
        basename = resource_info.get("basic", "name")
        dirname = dir_path
        path = os.path.join(dirname, basename)

        dict_ = {
            "name": resource_info.get("rspace", "name"),
            "uri": self.uri_from_path(path),
            "path": path,
        }

        entry: AnyRemoteEntry
        if resource_info.is_dir:
            dict_.update(
                {
                    "class": "Directory",
                }
            )
            entry = RemoteDirectory(**cast(RemoteDirectory, dict_))
        else:
            dict_.update(
                {
                    "class": "File",
                    "size": resource_info.size,
                    "ctime": self.to_dict_time(
                        datetime.datetime.fromisoformat(resource_info.get("rspace", "created")).astimezone(
                            datetime.timezone.utc
                        )
                    ),
                }
            )
            entry = RemoteFile(**cast(RemoteFile, dict_))

        return entry



import logging
import zlib
from dataclasses import (
    dataclass,
    fields,
)
from struct import unpack
from typing import (
    Dict,
    Optional,
)
from urllib.parse import (
    parse_qs,
    unquote,
    urlparse,
)

import requests
from typing_extensions import Unpack

from galaxy.files import OptionalUserContext
from galaxy.files.uris import validate_uri_access
from . import (
    BaseFilesSource,
    FilesSourceOptions,
    FilesSourceProperties,
    PluginKind,
)

log = logging.getLogger(__name__)


@dataclass
class FileExtractParameters:
    source: str
    header_offset: int
    compression_method: int
    compress_size: int


DOC_TEMPLATE = """
Remote ZIP archive extractor
---------------------------

This plugin extracts files from a public remotely hosted ZIP archive using byte-range requests.
The following parameters are required and assumed to be known in advance:

- `source`: The URL of the remote ZIP archive.
- `header_offset`: The offset of the file header in the archive.
- `compression_method`: The compression method used for the file.
- `compress_size`: The size of the compressed file.
"""


class RemoteZipFilesSource(BaseFilesSource):
    plugin_type = "remoteZip"
    plugin_kind = PluginKind.stock

    def __init__(self, **kwd: Unpack[FilesSourceProperties]):
        kwds: FilesSourceProperties = {
            "id": "extract",
            "label": "Remote ZIP extractor",
            "doc": DOC_TEMPLATE,
            "writable": False,
            "browsable": False,
        }
        kwds.update(kwd)
        props = self._parse_common_config_opts(kwds)
        self._props = props

    @property
    def _allowlist(self):
        return self._file_sources_config.fetch_url_allowlist

    def _realize_to(
        self,
        source_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        params = extract_query_parameters(source_path)
        file_extract_params = validate_params(params)
        validate_uri_access(
            file_extract_params.source,
            user_context.is_admin if user_context else False,
            self._allowlist or [],
        )
        stream_and_decompress(file_extract_params, native_path)

    def _write_from(
        self,
        target_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        raise NotImplementedError()

    def score_url_match(self, url: str):
        if url.startswith("zip://"):
            return len("zip://")
        else:
            return 0

    def _serialization_props(self, user_context: OptionalUserContext = None):
        effective_props = {}
        for key, val in self._props.items():
            effective_props[key] = self._evaluate_prop(val, user_context=user_context)
        return effective_props


def extract_query_parameters(url: str) -> Dict[str, str]:
    parsed_url = urlparse(url)
    query_params = parse_qs(parsed_url.query)
    return {key: value[0] for key, value in query_params.items()}


def validate_params(params: Dict[str, str]) -> FileExtractParameters:
    """Validates and converts the params dictionary to a FileExtractParameters instance."""

    required_fields = [field.name for field in fields(FileExtractParameters)]

    # Ensure all required fields are present
    for field in required_fields:
        if field not in params:
            raise ValueError(f"Missing required parameter: '{field}'")

    # Validate URL
    url = urlparse(unquote(params["source"]))
    if not url.scheme or not url.netloc:
        raise ValueError(f"Invalid URL provided: '{params['source']}'")

    # Validate integer parameters
    try:
        header_offset = int(params["header_offset"])
        compression_method = int(params["compression_method"])
        compress_size = int(params["compress_size"])
    except ValueError:
        raise ValueError("header_offset, compression_method, and compress_size must be integers.")

    return FileExtractParameters(
        source=params["source"],
        header_offset=header_offset,
        compression_method=compression_method,
        compress_size=compress_size,
    )


def get_final_url(url: str):
    """Follow redirects and return the final URL."""
    session = requests.Session()
    response = session.head(url, allow_redirects=True)
    return response.url


def get_range(url: str, start: int, length: int):
    """Fetch a specific byte range from a remote file."""
    headers = {"Range": f"bytes={start}-{start + length - 1}"}
    response = requests.get(url, headers=headers, stream=True)
    response.raise_for_status()
    return response


def calculate_file_data_offset(url, header_offset: int):
    """Determine the offset of the file data inside the ZIP archive."""
    stream_response = get_range(url, header_offset, 30)
    local_header_data = stream_response.content
    file_name_length, extra_field_length = unpack("<HH", local_header_data[26:30])
    return header_offset + 30 + file_name_length + extra_field_length


def stream_and_decompress(file_extract_params: FileExtractParameters, output_file: str):
    """Stream a file from a ZIP archive hosted remotely using byte-range requests."""
    final_url = get_final_url(file_extract_params.source)

    # Calculate actual start of file data
    file_data_offset = calculate_file_data_offset(final_url, file_extract_params.header_offset)

    # Stream and decompress file
    with get_range(final_url, file_data_offset, file_extract_params.compress_size) as response_stream:
        if file_extract_params.compression_method == 0:  # Stored (no compression)
            with open(output_file, "wb") as f:
                for chunk in response_stream.iter_content(chunk_size=8192):
                    f.write(chunk)
            return output_file

        if file_extract_params.compression_method == 8:  # Deflate
            decompressor = zlib.decompressobj(-zlib.MAX_WBITS)
        else:
            raise ValueError(f"Unsupported compression method: {file_extract_params.compression_method}")

        with open(output_file, "wb") as f:
            for chunk in response_stream.iter_content(chunk_size=8192):
                if chunk:
                    f.write(decompressor.decompress(chunk))

            # Flush remaining decompressed data
            f.write(decompressor.flush())

        return output_file


__all__ = ("RemoteZipFilesSource",)



"""Static Galaxy file sources - ftp and libraries."""

from typing import (
    Optional,
)

from typing_extensions import Unpack

from galaxy.files.sources import PluginKind
from .posix import (
    PosixFilesSource,
    PosixFilesSourceProperties,
)


class UserFtpFilesSource(PosixFilesSource):
    plugin_type = "gxftp"
    plugin_kind = PluginKind.stock

    def __init__(self, **kwd: Unpack[PosixFilesSourceProperties]):
        posix_kwds: PosixFilesSourceProperties = dict(
            id="_ftp",
            root="${user.ftp_dir}",
            label="FTP Directory",
            doc="Galaxy User's FTP Directory",
            writable=True,
        )
        posix_kwds.update(kwd)
        if "delete_on_realize" not in posix_kwds:
            file_sources_config = kwd["file_sources_config"]
            posix_kwds["delete_on_realize"] = file_sources_config.ftp_upload_purge
        super().__init__(**posix_kwds)

    def get_prefix(self) -> Optional[str]:
        return None

    def get_scheme(self) -> str:
        return "gxftp"


class LibraryImportFilesSource(PosixFilesSource):
    plugin_type = "gximport"
    plugin_kind = PluginKind.stock

    def __init__(self, **kwd: Unpack[PosixFilesSourceProperties]):
        posix_kwds: PosixFilesSourceProperties = dict(
            id="_import",
            root="${config.library_import_dir}",
            label="Library Import Directory",
            doc="Galaxy's library import directory",
        )
        posix_kwds.update(kwd)
        super().__init__(**posix_kwds)

    def get_prefix(self) -> Optional[str]:
        return None

    def get_scheme(self) -> str:
        return "gximport"


class UserLibraryImportFilesSource(PosixFilesSource):
    plugin_type = "gxuserimport"
    plugin_kind = PluginKind.stock

    def __init__(self, **kwd: Unpack[PosixFilesSourceProperties]):
        posix_kwds: PosixFilesSourceProperties = dict(
            id="_userimport",
            root="${config.user_library_import_dir}/${user.email}",
            label="Library User Import Directory",
            doc="Galaxy's user library import directory",
        )
        posix_kwds.update(kwd)
        super().__init__(**posix_kwds)

    def get_prefix(self) -> Optional[str]:
        return None

    def get_scheme(self) -> str:
        return "gxuserimport"


__all__ = ("UserFtpFilesSource", "LibraryImportFilesSource", "UserLibraryImportFilesSource")



from typing import (
    cast,
    List,
    Optional,
    Type,
    TYPE_CHECKING,
)

from galaxy.util.config_parsers import parse_allowlist_ips
from galaxy.util.plugin_config import (
    load_plugins,
    PluginConfigSource,
    plugins_dict,
)

if TYPE_CHECKING:
    from galaxy.files.sources import BaseFilesSource


class FileSourcePluginsConfig:
    symlink_allowlist: List[str]
    fetch_url_allowlist: List[str]
    library_import_dir: Optional[str]
    user_library_import_dir: Optional[str]
    ftp_upload_dir: Optional[str]
    ftp_upload_purge: bool
    tmp_dir: Optional[str]
    webdav_use_temp_files: Optional[bool]
    listings_expiry_time: Optional[int]

    def __init__(
        self,
        symlink_allowlist=None,
        fetch_url_allowlist=None,
        library_import_dir=None,
        user_library_import_dir=None,
        ftp_upload_dir=None,
        ftp_upload_purge=True,
        tmp_dir=None,
        webdav_use_temp_files=None,
        listings_expiry_time=None,
    ):
        symlink_allowlist = symlink_allowlist or []
        fetch_url_allowlist = fetch_url_allowlist or []
        self.symlink_allowlist = symlink_allowlist
        self.fetch_url_allowlist = fetch_url_allowlist
        self.library_import_dir = library_import_dir
        self.user_library_import_dir = user_library_import_dir
        self.ftp_upload_dir = ftp_upload_dir
        self.ftp_upload_purge = ftp_upload_purge
        self.tmp_dir = tmp_dir
        self.webdav_use_temp_files = webdav_use_temp_files
        self.listings_expiry_time = listings_expiry_time

    @staticmethod
    def from_app_config(config):
        # Formalize what we read in from config to create a more clear interface
        # for this component.
        kwds = {}
        kwds["symlink_allowlist"] = config.user_library_import_symlink_allowlist
        kwds["fetch_url_allowlist"] = [str(ip) for ip in config.fetch_url_allowlist_ips]
        kwds["library_import_dir"] = config.library_import_dir
        kwds["user_library_import_dir"] = config.user_library_import_dir
        kwds["ftp_upload_dir"] = config.ftp_upload_dir
        kwds["ftp_upload_purge"] = config.ftp_upload_purge
        kwds["tmp_dir"] = config.file_source_temp_dir
        kwds["webdav_use_temp_files"] = config.file_source_webdav_use_temp_files
        kwds["listings_expiry_time"] = config.file_source_listings_expiry_time

        return FileSourcePluginsConfig(**kwds)

    def to_dict(self):
        return {
            "symlink_allowlist": self.symlink_allowlist,
            "fetch_url_allowlist": self.fetch_url_allowlist,
            "library_import_dir": self.library_import_dir,
            "user_library_import_dir": self.user_library_import_dir,
            "ftp_upload_dir": self.ftp_upload_dir,
            "ftp_upload_purge": self.ftp_upload_purge,
            "tmp_dir": self.tmp_dir,
            "webdav_use_temp_files": self.webdav_use_temp_files,
            "listings_expiry_time": self.listings_expiry_time,
        }

    @staticmethod
    def from_dict(as_dict):
        return FileSourcePluginsConfig(
            symlink_allowlist=as_dict["symlink_allowlist"],
            fetch_url_allowlist=parse_allowlist_ips(as_dict["fetch_url_allowlist"]),
            library_import_dir=as_dict["library_import_dir"],
            user_library_import_dir=as_dict["user_library_import_dir"],
            ftp_upload_dir=as_dict["ftp_upload_dir"],
            ftp_upload_purge=as_dict["ftp_upload_purge"],
            # Always provided for new jobs, remove in 25.0
            tmp_dir=as_dict.get("tmp_dir"),
            webdav_use_temp_files=as_dict.get("webdav_use_temp_files"),
            listings_expiry_time=as_dict.get("listings_expiry_time"),
        )


class FileSourcePluginLoader:

    def __init__(self):
        self._plugin_classes = self._file_source_plugins_dict()

    def _file_source_plugins_dict(self):
        import galaxy.files.sources

        return plugins_dict(galaxy.files.sources, "plugin_type")

    def get_plugin_type_class(self, plugin_type: str) -> Type["BaseFilesSource"]:
        return cast(Type["BaseFilesSource"], self._plugin_classes[plugin_type])

    def load_plugins(
        self, plugin_source: PluginConfigSource, file_source_plugin_config: FileSourcePluginsConfig
    ) -> List["BaseFilesSource"]:
        extra_kwds = {
            "file_sources_config": file_source_plugin_config,
        }
        return load_plugins(
            self._plugin_classes,
            plugin_source,
            extra_kwds,
            dict_to_list_key="id",
        )


__all__ = (
    "FileSourcePluginLoader",
    "FileSourcePluginsConfig",
)



from .manager import ConfiguredFileSourceTemplates
from .models import (
    FileSourceConfiguration,
    FileSourceTemplate,
    FileSourceTemplateSummaries,
    FileSourceTemplateType,
    get_oauth2_config,
    get_oauth2_config_or_none,
    template_to_configuration,
)

__all__ = (
    "ConfiguredFileSourceTemplates",
    "FileSourceConfiguration",
    "FileSourceTemplate",
    "FileSourceTemplateSummaries",
    "FileSourceTemplateType",
    "get_oauth2_config",
    "get_oauth2_config_or_none",
    "template_to_configuration",
)



import datetime
import json
import re
import urllib.request
from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Tuple,
)
from urllib.parse import quote

from typing_extensions import (
    Literal,
    TypedDict,
    Unpack,
)

from galaxy.exceptions import (
    AuthenticationRequired,
    MessageException,
)
from galaxy.files import OptionalUserContext
from galaxy.files.sources import (
    AnyRemoteEntry,
    DEFAULT_PAGE_LIMIT,
    DEFAULT_SCHEME,
    Entry,
    EntryData,
    FilesSourceOptions,
    RemoteDirectory,
    RemoteFile,
)
from galaxy.files.sources._rdm import (
    ContainerAndFileIdentifier,
    RDMFilesSource,
    RDMFilesSourceProperties,
    RDMRepositoryInteractor,
)
from galaxy.util import (
    DEFAULT_SOCKET_TIMEOUT,
    get_charset_from_http_headers,
    requests,
    stream_to_open_named_file,
)

AccessStatus = Literal["public", "restricted"]


class ResourceType(TypedDict):
    id: str


class RecordAccess(TypedDict):
    record: AccessStatus
    files: AccessStatus


class RecordFiles(TypedDict):
    enabled: bool


class IdentifierEntry(TypedDict):
    scheme: str
    identifier: str


class AffiliationEntry(TypedDict):
    id: str
    name: str


class RecordPersonOrOrg(TypedDict):
    family_name: str
    given_name: str
    type: Literal["personal", "organizational"]
    name: str
    identifiers: List[IdentifierEntry]


class Creator(TypedDict):
    person_or_org: RecordPersonOrOrg
    affiliations: Optional[List[AffiliationEntry]]


class RecordMetadata(TypedDict):
    title: str
    resource_type: ResourceType
    publication_date: str
    creators: List[Creator]


class RecordLinks(TypedDict):
    self: str
    self_html: str
    self_iiif_manifest: str
    self_iiif_sequence: str
    files: str
    record: str
    record_html: str
    publish: str
    review: str
    versions: str
    access_links: str
    reserve_doi: str


class InvenioRecord(TypedDict):
    id: str
    title: str
    created: str
    updated: str
    resource_type: ResourceType
    publication_date: str
    access: RecordAccess
    files: RecordFiles
    metadata: RecordMetadata
    links: RecordLinks


class InvenioRDMFilesSource(RDMFilesSource):
    """A files source for Invenio turn-key research data management repository.

    In Invenio a "Record" represents what we refer to as container in the rdm base class
    """

    plugin_type = "inveniordm"
    supports_pagination = True
    supports_search = True
    rdm_scheme = "invenio"

    def __init__(self, **kwd: Unpack[RDMFilesSourceProperties]):
        super().__init__(**kwd)
        self._scheme_regex = re.compile(rf"^{self.get_scheme()}?://{self.id}|^{DEFAULT_SCHEME}://{self.id}")
        self.repository: InvenioRepositoryInteractor

    def get_scheme(self) -> str:
        return self.scheme if self.scheme and self.scheme != DEFAULT_SCHEME else self.rdm_scheme

    def score_url_match(self, url: str) -> int:
        if match := self._scheme_regex.match(url):
            return match.span()[1]
        else:
            return 0

    def to_relative_path(self, url: str) -> str:
        legacy_uri_root = f"{DEFAULT_SCHEME}://{self.id}"
        if url.startswith(legacy_uri_root):
            return url[len(legacy_uri_root) :]
        else:
            return super().to_relative_path(url)

    def get_repository_interactor(self, repository_url: str) -> RDMRepositoryInteractor:
        return InvenioRepositoryInteractor(repository_url, self)

    def parse_path(self, source_path: str, container_id_only: bool = False) -> ContainerAndFileIdentifier:
        """Parses the given source path and returns the record_id and filename.

        The source path must have the format '/<record_id>/<file_name>'.
        If container_id_only is True, the source path must have the format '/<record_id>' and and an empty filename will be returned.
        """

        def get_error_msg(details: str) -> str:
            return f"Invalid source path: '{source_path}'. Expected format: '{expected_format}'. {details}"

        expected_format = "/<record_id>"
        if not source_path.startswith("/"):
            raise ValueError(get_error_msg("Must start with '/'."))
        parts = source_path[1:].split("/", 2)
        if container_id_only:
            if len(parts) != 1:
                raise ValueError(get_error_msg("Please provide the record_id only."))
            return ContainerAndFileIdentifier(container_id=parts[0], file_identifier="")
        expected_format = "/<record_id>/<file_name>"
        if len(parts) < 2:
            raise ValueError(get_error_msg("Please provide both the record_id and file_name."))
        if len(parts) > 2:
            # TODO: This causes downloads to crash if the filename contains a slash
            raise ValueError(get_error_msg("Too many parts. Please provide the record_id and file_name only."))
        record_id, file_name = parts
        return ContainerAndFileIdentifier(container_id=record_id, file_identifier=file_name)

    def get_container_id_from_path(self, source_path: str) -> str:
        return self.parse_path(source_path, container_id_only=True).container_id

    def _list(
        self,
        path="/",
        recursive=True,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        sort_by: Optional[str] = None,
    ) -> Tuple[List[AnyRemoteEntry], int]:
        writeable = opts and opts.writeable or False
        is_root_path = path == "/"
        if is_root_path:
            records, total_hits = self.repository.get_file_containers(
                writeable, user_context, limit=limit, offset=offset, query=query
            )
            return cast(List[AnyRemoteEntry], records), total_hits
        record_id = self.get_container_id_from_path(path)
        files = self.repository.get_files_in_container(record_id, writeable, user_context)
        return cast(List[AnyRemoteEntry], files), len(files)

    def _create_entry(
        self,
        entry_data: EntryData,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ) -> Entry:
        public_name = self.get_public_name(user_context) or "No name"
        record = self.repository.create_draft_file_container(entry_data["name"], public_name, user_context=user_context)
        record_id = record.get("id")
        record_id = str(record_id) if record_id else None
        if not record_id:
            raise Exception("Failed to create record.")
        uri = self.repository.to_plugin_uri(record_id=record_id)
        name = record.get("title") or "Untitled"
        if not isinstance(name, str):
            raise Exception("Failed to get record title.")
        links = record.get("links")
        if not links or not isinstance(links, dict):
            raise Exception("Failed to get record links.")
        external_link = links.get("self_html")
        if not external_link or not isinstance(external_link, str):
            raise Exception("Failed to get record link.")
        return {
            "uri": uri,
            "name": name,
            "external_link": external_link,
        }

    def _realize_to(
        self,
        source_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        # TODO: user_context is always None here when called from a data fetch.
        # This prevents downloading files that require authentication even if the user provided a token.
        record_id, filename = self.parse_path(source_path)
        self.repository.download_file_from_container(record_id, filename, native_path, user_context=user_context)

    def _write_from(
        self,
        target_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        record_id, filename = self.parse_path(target_path)
        self.repository.upload_file_to_draft_container(record_id, filename, native_path, user_context=user_context)


class InvenioRepositoryInteractor(RDMRepositoryInteractor):
    """In Invenio a "Record" represents what we refer to as container in the rdm base class"""

    @property
    def records_url(self) -> str:
        return f"{self.repository_url}/api/records"

    @property
    def user_records_url(self) -> str:
        return f"{self.repository_url}/api/user/records"

    def to_plugin_uri(self, record_id: str, filename: Optional[str] = None) -> str:
        return f"{self.plugin.get_uri_root()}/{record_id}{f'/{filename}' if filename else ''}"

    def get_file_containers(
        self,
        writeable: bool,
        user_context: OptionalUserContext = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        sort_by: Optional[str] = None,
    ) -> Tuple[List[RemoteDirectory], int]:
        """Gets the records in the repository and returns the total count of records."""
        params: Dict[str, Any] = {}
        request_url = self.records_url
        if writeable:
            # Only draft records owned by the user can be written to.
            params["is_published"] = "false"
            request_url = self.user_records_url
        size, page = self._to_size_page(limit, offset)
        params["size"] = size
        params["page"] = page
        if query:
            params["q"] = query
            params["sort"] = "bestmatch"
        response_data = self._get_response(user_context, request_url, params=params)
        total_hits = response_data["hits"]["total"]
        return self._get_records_from_response(response_data), total_hits

    def _to_size_page(self, limit: Optional[int], offset: Optional[int]) -> Tuple[Optional[int], Optional[int]]:
        if limit is None and offset is None:
            return None, None
        size = limit or DEFAULT_PAGE_LIMIT
        page = (offset or 0) // size + 1
        return size, page

    def get_files_in_container(
        self,
        container_id: str,
        writeable: bool,
        user_context: OptionalUserContext = None,
        query: Optional[str] = None,
    ) -> List[RemoteFile]:
        conditionally_draft = "/draft" if writeable else ""
        request_url = f"{self.records_url}/{container_id}{conditionally_draft}/files"
        response_data = self._get_response(user_context, request_url)
        return self._get_record_files_from_response(container_id, response_data)

    def create_draft_file_container(
        self, title: str, public_name: str, user_context: OptionalUserContext = None
    ) -> RemoteDirectory:
        today = datetime.date.today().isoformat()
        creator = self._get_creator_from_public_name(public_name)
        create_record_request = {
            "files": {"enabled": True},
            "metadata": {
                "title": title,
                "publication_date": today,
                "resource_type": {"id": "dataset"},
                "creators": [
                    creator,
                ],
            },
        }

        headers = self._get_request_headers(user_context, auth_required=True)
        response = requests.post(self.records_url, json=create_record_request, headers=headers)
        self._ensure_response_has_expected_status_code(response, 201)
        record = response.json()
        record["title"] = self._get_record_title(record)
        return record

    def upload_file_to_draft_container(
        self,
        record_id: str,
        filename: str,
        file_path: str,
        user_context: OptionalUserContext = None,
    ):
        record = self._get_draft_record(record_id, user_context=user_context)
        upload_file_url = record["links"]["files"]
        headers = self._get_request_headers(user_context, auth_required=True)

        # Add file metadata entry
        response = requests.post(upload_file_url, json=[{"key": filename}], headers=headers)
        self._ensure_response_has_expected_status_code(response, 201)

        # Upload file content
        entries = response.json()["entries"]
        file_entry = next(entry for entry in entries if entry["key"] == filename)
        upload_file_content_url = file_entry["links"]["content"]
        commit_file_upload_url = file_entry["links"]["commit"]
        with open(file_path, "rb") as file:
            response = requests.put(upload_file_content_url, data=file, headers=headers)
            self._ensure_response_has_expected_status_code(response, 200)

        # Commit file upload
        response = requests.post(commit_file_upload_url, headers=headers)
        self._ensure_response_has_expected_status_code(response, 200)

    def download_file_from_container(
        self,
        container_id: str,
        file_identifier: str,
        file_path: str,
        user_context: OptionalUserContext = None,
    ):
        download_file_content_url = self._get_download_file_url(container_id, file_identifier, user_context)
        headers = {}
        if self._is_api_url(download_file_content_url):
            # pass the token as a header only when using the API
            headers = self._get_request_headers(user_context)
        try:
            req = urllib.request.Request(download_file_content_url, headers=headers)
            with urllib.request.urlopen(req, timeout=DEFAULT_SOCKET_TIMEOUT) as page:
                f = open(file_path, "wb")
                return stream_to_open_named_file(
                    page, f.fileno(), file_path, source_encoding=get_charset_from_http_headers(page.headers)
                )
        except urllib.error.HTTPError as e:
            # TODO: We can only download files from published records for now
            if e.code in [401, 403, 404]:
                raise Exception(
                    f"Cannot download file '{file_identifier}' from record '{container_id}'. Please make sure the record exists and it is public."
                )

    def _get_download_file_url(self, record_id: str, filename: str, user_context: OptionalUserContext = None):
        """Get the URL to download a file from a record.

        This method is used to download files from both published and draft records that are accessible by the user.
        """
        file_details_url = self._get_file_details_url(record_id, filename)
        if self._is_published_record(record_id, user_context):
            return self._file_url_to_download_url(file_details_url)
        if self._is_draft_record(record_id, user_context):
            draft_download_url = f"{self._to_draft_url(file_details_url)}/content"
            return draft_download_url
        raise MessageException(
            f"Cannot download file '{filename}' from record '{record_id}'. The record is not accessible or does not exist."
        )

    def _is_api_url(self, url: str) -> bool:
        return "/api/" in url

    def _to_draft_url(self, url: str) -> str:
        return url.replace("/files/", "/draft/files/")

    def _is_draft_record(self, record_id: str, user_context: OptionalUserContext = None):
        request_url = self._get_draft_record_url(record_id)
        headers = self._get_request_headers(user_context)
        response = requests.head(request_url, headers=headers)
        return response.status_code == 200

    def _is_published_record(self, record_id: str, user_context: OptionalUserContext = None):
        request_url = self._get_record_url(record_id)
        headers = self._get_request_headers(user_context)
        response = requests.head(request_url, headers=headers)
        return response.status_code == 200

    def _get_record_url(self, record_id: str):
        return f"{self.records_url}/{record_id}"

    def _get_draft_record_url(self, record_id: str):
        return f"{self._get_record_url(record_id)}/draft"

    def _get_file_details_url(self, record_id: str, filename: str):
        return f"{self._get_record_url(record_id)}/files/{quote(filename)}"

    def _file_url_to_download_url(self, file_url: str) -> str:
        # Downloading through the API is only supported for local files and depends on how
        # the InvenioRDM instance file storage is configured.
        # So this is the most reliable way to download files for now.
        return f"{file_url.replace('/api', '')}?download=1"

    def _get_draft_record(self, record_id: str, user_context: OptionalUserContext = None):
        request_url = self._get_draft_record_url(record_id)
        draft_record = self._get_response(user_context, request_url)
        return draft_record

    def _get_records_from_response(self, response: dict) -> List[RemoteDirectory]:
        records = response["hits"]["hits"]
        rval: List[RemoteDirectory] = []
        for record in records:
            uri = self.to_plugin_uri(record_id=record["id"])
            path = self.plugin.to_relative_path(uri)
            name = self._get_record_title(record)
            rval.append(
                {
                    "class": "Directory",
                    "name": name,
                    "uri": uri,
                    "path": path,
                }
            )
        return rval

    def _get_record_title(self, record: InvenioRecord) -> str:
        title = record.get("title")
        if not title and "metadata" in record:
            title = record["metadata"].get("title")
        return title or "No title"

    def _get_record_files_from_response(self, record_id: str, response: dict) -> List[RemoteFile]:
        files_enabled = response.get("enabled", False)
        if not files_enabled:
            return []
        entries = response["entries"]
        rval: List[RemoteFile] = []
        for entry in entries:
            if entry.get("status") == "completed":
                uri = self.to_plugin_uri(record_id=record_id, filename=entry["key"])
                path = self.plugin.to_relative_path(uri)
                rval.append(
                    {
                        "class": "File",
                        "name": entry["key"],
                        "size": entry["size"],
                        "ctime": entry["created"],
                        "uri": uri,
                        "path": path,
                    }
                )
        return rval

    def _get_creator_from_public_name(self, public_name: Optional[str] = None) -> Creator:
        given_name = "Anonymous"
        family_name = "Galaxy User"
        if public_name:
            tokens = public_name.split(", ")
            if len(tokens) == 2:
                family_name = tokens[0]
                given_name = tokens[1]
            else:
                given_name = public_name
        return {
            "person_or_org": {
                "name": f"{given_name} {family_name}",
                "family_name": family_name,
                "given_name": given_name,
                "type": "personal",
                "identifiers": [],
            },
            "affiliations": [],
        }

    def _get_response(
        self,
        user_context: OptionalUserContext,
        request_url: str,
        params: Optional[Dict[str, Any]] = None,
        auth_required: bool = False,
    ) -> dict:
        headers = self._get_request_headers(user_context, auth_required)
        response = requests.get(request_url, params=params, headers=headers)
        self._ensure_response_has_expected_status_code(response, 200)
        return response.json()

    def _get_request_headers(self, user_context: OptionalUserContext, auth_required: bool = False):
        token = self.plugin.get_authorization_token(user_context)
        headers = {"Authorization": f"Bearer {token}"} if token else {}
        if auth_required and token is None:
            self._raise_auth_required()
        return headers

    def _ensure_response_has_expected_status_code(self, response, expected_status_code: int):
        if response.status_code != expected_status_code:
            if response.status_code == 403:
                self._raise_auth_required()
            error_message = self._get_response_error_message(response)
            raise Exception(
                f"Request to {response.url} failed with status code {response.status_code}: {error_message}"
            )

    def _raise_auth_required(self):
        raise AuthenticationRequired(
            f"Please provide a personal access token in your user's preferences for '{self.plugin.label}'"
        )

    def _get_response_error_message(self, response):
        response_json = response.json()
        error_message = response_json.get("message") if response.status_code == 400 else response.text
        errors = response_json.get("errors", [])
        for error in errors:
            error_message += f"\n{json.dumps(error)}"
        return error_message


__all__ = ("InvenioRDMFilesSource",)



try:
    from webdavfs.webdavfs import WebDAVFS
except ImportError:
    WebDAVFS = None

import tempfile
from typing import (
    cast,
    Optional,
    Union,
)

from typing_extensions import NotRequired

from . import (
    FilesSourceOptions,
    FilesSourceProperties,
)
from ._pyfilesystem2 import PyFilesystem2FilesSource


class WebDavFilesSourceProperties(FilesSourceProperties, total=False):
    use_temp_files: NotRequired[Optional[bool]]
    temp_path: NotRequired[Optional[str]]


class WebDavFilesSource(PyFilesystem2FilesSource):
    plugin_type = "webdav"
    required_module = WebDAVFS
    required_package = "fs.webdavfs"
    allow_key_error_on_empty_directories = True

    def _open_fs(self, user_context=None, opts: Optional[FilesSourceOptions] = None):
        props = cast(WebDavFilesSourceProperties, self._serialization_props(user_context))
        file_sources_config = self._file_sources_config
        use_temp_files = props.pop("use_temp_files", None)
        if use_temp_files is None and file_sources_config and file_sources_config.webdav_use_temp_files is not None:
            use_temp_files = file_sources_config.webdav_use_temp_files
        if use_temp_files is None:
            # Default to True to avoid memory issues with large files.
            use_temp_files = True

        if use_temp_files:
            temp_path = props.get("temp_path")
            if temp_path is None and file_sources_config and file_sources_config.tmp_dir:
                temp_path = file_sources_config.tmp_dir
            if temp_path is None:
                temp_path = tempfile.mkdtemp(prefix="webdav_")
            props["temp_path"] = temp_path
        props["use_temp_files"] = use_temp_files
        extra_props: Union[FilesSourceProperties, dict] = opts.extra_props or {} if opts else {}
        handle = WebDAVFS(**{**props, **extra_props})
        return handle


__all__ = ("WebDavFilesSource",)



import abc
import functools
import logging
import os
from typing import (
    ClassVar,
    List,
    Optional,
    Tuple,
    Type,
)

import fs
import fs.errors
from fs.base import FS
from typing_extensions import Unpack

from galaxy.exceptions import (
    AuthenticationRequired,
    MessageException,
)
from galaxy.files import OptionalUserContext
from . import (
    AnyRemoteEntry,
    BaseFilesSource,
    DEFAULT_PAGE_LIMIT,
    FilesSourceOptions,
    FilesSourceProperties,
)

log = logging.getLogger(__name__)

PACKAGE_MESSAGE = "FilesSource plugin is missing required Python PyFilesystem2 plugin package [%s]"


class PyFilesystem2FilesSource(BaseFilesSource):
    required_module: ClassVar[Optional[Type[FS]]]
    required_package: ClassVar[str]
    supports_pagination = True
    supports_search = True
    allow_key_error_on_empty_directories = False  # work around a bug in webdav

    def __init__(self, **kwd: Unpack[FilesSourceProperties]):
        if self.required_module is None:
            raise Exception(PACKAGE_MESSAGE % self.required_package)
        props = self._parse_common_config_opts(kwd)
        self._props = props

    @abc.abstractmethod
    def _open_fs(self, user_context: OptionalUserContext = None, opts: Optional[FilesSourceOptions] = None) -> FS:
        """Subclasses must instantiate a PyFilesystem2 handle for this file system."""

    def _list(
        self,
        path="/",
        recursive=False,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        sort_by: Optional[str] = None,
    ) -> Tuple[List[AnyRemoteEntry], int]:
        """Return dictionary of 'Directory's and 'File's."""
        try:
            with self._open_fs(user_context=user_context, opts=opts) as h:
                if recursive:
                    recursive_result: List[AnyRemoteEntry] = []
                    try:
                        for p, dirs, files in h.walk(path, namespaces=["details"]):
                            to_dict = functools.partial(self._resource_info_to_dict, p)
                            recursive_result.extend(map(to_dict, dirs))
                            recursive_result.extend(map(to_dict, files))
                    except KeyError:
                        if not self.allow_key_error_on_empty_directories:
                            raise
                    return recursive_result, len(recursive_result)
                else:
                    page = self._to_page(limit, offset)
                    filter = self._query_to_filter(query)
                    count = self._get_total_matches_count(h, path, filter)
                    result = h.filterdir(path, namespaces=["details"], page=page, files=filter, dirs=filter)
                    to_dict = functools.partial(self._resource_info_to_dict, path)
                    return list(map(to_dict, result)), count
        except fs.errors.PermissionDenied as e:
            raise AuthenticationRequired(
                f"Permission Denied. Reason: {e}. Please check your credentials in your preferences for {self.label}."
            )
        except fs.errors.FSError as e:
            raise MessageException(f"Problem listing file source path {path}. Reason: {e}") from e

    def _get_total_matches_count(self, fs: FS, path: str, filter: Optional[List[str]] = None) -> int:
        return sum(1 for _ in fs.filterdir(path, namespaces=["basic"], files=filter, dirs=filter))

    def _to_page(self, limit: Optional[int] = None, offset: Optional[int] = None) -> Optional[Tuple[int, int]]:
        if limit is None and offset is None:
            return None
        limit = limit or DEFAULT_PAGE_LIMIT
        start = offset or 0
        end = start + limit
        return (start, end)

    def _query_to_filter(self, query: Optional[str]) -> Optional[List[str]]:
        if not query:
            return None
        return [f"*{query}*"]

    def _realize_to(
        self,
        source_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        with open(native_path, "wb") as write_file:
            self._open_fs(user_context=user_context, opts=opts).download(source_path, write_file)

    def _write_from(
        self,
        target_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        with open(native_path, "rb") as read_file:
            openfs = self._open_fs(user_context=user_context, opts=opts)
            dirname = fs.path.dirname(target_path)
            if not openfs.isdir(dirname):
                openfs.makedirs(dirname)
            openfs.upload(target_path, read_file)

    def _resource_info_to_dict(self, dir_path, resource_info) -> AnyRemoteEntry:
        name = resource_info.name
        path = os.path.join(dir_path, name)
        uri = self.uri_from_path(path)
        if resource_info.is_dir:
            return {"class": "Directory", "name": name, "uri": uri, "path": path}
        else:
            created = resource_info.created
            return {
                "class": "File",
                "name": name,
                "size": resource_info.size,
                "ctime": self.to_dict_time(created),
                "uri": uri,
                "path": path,
            }

    def _serialization_props(self, user_context: OptionalUserContext = None):
        effective_props = {}
        for key, val in self._props.items():
            effective_props[key] = self._evaluate_prop(val, user_context=user_context)
        return effective_props



import logging
from typing import (
    List,
    NamedTuple,
    Optional,
    Tuple,
)

from typing_extensions import Unpack

from galaxy.files import OptionalUserContext
from galaxy.files.sources import (
    BaseFilesSource,
    FilesSourceProperties,
    PluginKind,
    RemoteDirectory,
    RemoteFile,
)

log = logging.getLogger(__name__)


class RDMFilesSourceProperties(FilesSourceProperties):
    token: str
    public_name: str


class ContainerAndFileIdentifier(NamedTuple):
    """The file_identifier could be a filename or a file_id."""

    container_id: str
    file_identifier: str


class RDMRepositoryInteractor:
    """Base class for interacting with an external RDM repository.

    This class is not intended to be used directly, but rather to be subclassed
    by file sources that interact with RDM repositories.

    Different RDM repositories use different terminology. Also they use the same term for different things.
    To prevent confusion, we use the term "container" in the base repository.
    This is an abstract term for the entity that contains multiple files.
    """

    def __init__(self, repository_url: str, plugin: "RDMFilesSource"):
        self._repository_url = repository_url
        self._plugin = plugin

    @property
    def plugin(self) -> "RDMFilesSource":
        """Returns the plugin associated with this repository interactor."""
        return self._plugin

    @property
    def repository_url(self) -> str:
        """Returns the base URL of the repository.

        Example: https://zenodo.org
        """
        return self._repository_url

    def to_plugin_uri(self, container_id: str, filename: Optional[str] = None) -> str:
        """Creates a valid plugin URI to reference the given container_id.

        If a filename is provided, the URI will reference the specific file in the container."""
        raise NotImplementedError()

    def get_file_containers(
        self,
        writeable: bool,
        user_context: OptionalUserContext = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        sort_by: Optional[str] = None,
    ) -> Tuple[List[RemoteDirectory], int]:
        """Returns the list of file containers in the repository and the total count containers.

        If writeable is True, only containers that the user can write to will be returned.
        The user_context might be required to authenticate the user in the repository.
        """
        raise NotImplementedError()

    def get_files_in_container(
        self,
        container_id: str,
        writeable: bool,
        user_context: OptionalUserContext = None,
        query: Optional[str] = None,
    ) -> List[RemoteFile]:
        """Returns the list of files of a file container.

        If writeable is True, we are signaling that the user intends to write to the container.
        """
        raise NotImplementedError()

    def create_draft_file_container(self, title: str, public_name: str, user_context: OptionalUserContext = None):
        """Creates a draft file container in the repository with basic metadata.

        The metadata is usually just the title of the container and the user that created it.
        Some plugins might also provide additional metadata defaults in the user settings."""
        raise NotImplementedError()

    def upload_file_to_draft_container(
        self,
        container_id: str,
        filename: str,
        file_path: str,
        user_context: OptionalUserContext = None,
    ) -> None:
        """Uploads a file with the provided filename (from file_path) to a draft container with the given container_id.

        The draft container must have been created in advance with the `create_draft_file_container` method.

        The file must exist in the file system at the given file_path.
        The user_context might be required to authenticate the user in the repository.
        """
        raise NotImplementedError()

    def download_file_from_container(
        self,
        container_id: str,
        file_identifier: str,
        file_path: str,
        user_context: OptionalUserContext = None,
    ) -> None:
        """Downloads a file with the provided filename from the container with the given container_id.

        The file will be downloaded to the file system at the given file_path.
        The user_context might be required to authenticate the user in the repository if the
        file is not publicly available.
        """
        raise NotImplementedError()


class RDMFilesSource(BaseFilesSource):
    """Base class for Research Data Management (RDM) file sources.

    This class is not intended to be used directly, but rather to be subclassed
    by file sources that interact with RDM repositories.

    A RDM file source is similar to a regular file source, but instead of tree of
    files and directories, it provides a (one level) list of containers
    that can contain only files (no subdirectories).

    In addition, RDM file sources might need to create a new container in advance in the
    repository, and then upload a file to it. This is done by calling the `_create_entry` method.
    """

    plugin_kind = PluginKind.rdm

    def __init__(self, **kwd: Unpack[RDMFilesSourceProperties]):
        props = self._parse_common_config_opts(kwd)
        self.url = props.get("url")
        if not self.url:
            raise Exception("URL for RDM repository must be provided in configuration")
        self._props = props
        self._repository_interactor = self.get_repository_interactor(self.url)

    @property
    def repository(self) -> RDMRepositoryInteractor:
        return self._repository_interactor

    def get_url(self) -> Optional[str]:
        return self.url

    def get_repository_interactor(self, repository_url: str) -> RDMRepositoryInteractor:
        """Returns an interactor compatible with the given repository URL.

        This must be implemented by subclasses."""
        raise NotImplementedError()

    def parse_path(self, source_path: str, container_id_only: bool = False) -> ContainerAndFileIdentifier:
        """Parses the given source path and returns the container_id and filename.

        If container_id_only is True, an empty filename will be returned.

        This must be implemented by subclasses."""
        raise NotImplementedError()

    def get_container_id_from_path(self, source_path: str) -> str:
        raise NotImplementedError()

    def _serialization_props(self, user_context: OptionalUserContext = None):
        effective_props = {}
        for key, val in self._props.items():
            effective_props[key] = self._evaluate_prop(val, user_context=user_context)
        return effective_props

    def get_authorization_token(self, user_context: OptionalUserContext) -> Optional[str]:
        effective_props = self._serialization_props(user_context)
        return effective_props.get("token")

    def get_public_name(self, user_context: OptionalUserContext) -> Optional[str]:
        effective_props = self._serialization_props(user_context)
        return effective_props.get("public_name")



import abc
import os
import time
from dataclasses import (
    dataclass,
    field,
)
from enum import Enum
from typing import (
    Any,
    ClassVar,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TYPE_CHECKING,
    Union,
)

from typing_extensions import (
    Literal,
    NotRequired,
    TypedDict,
)

from galaxy.exceptions import (
    ConfigurationError,
    ItemAccessibilityException,
    RequestParameterInvalidException,
)
from galaxy.files.plugins import FileSourcePluginsConfig
from galaxy.util.bool_expressions import (
    BooleanExpressionEvaluator,
    TokenContainedEvaluator,
)
from galaxy.util.template import fill_template

DEFAULT_SCHEME = "gxfiles"
DEFAULT_WRITABLE = False
DEFAULT_PAGE_LIMIT = 25

if TYPE_CHECKING:
    from galaxy.files import (
        FileSourcesUserContext,
        OptionalUserContext,
    )


class PluginKind(str, Enum):
    """Enum to distinguish between different kinds or categories of plugins."""

    rfs = "rfs"
    """Remote File System

    A remote file system is a file source that is backed by a remote file system
    that is accessible via a URI. Examples include FTP, SFTP, S3, etc.
    """

    drs = "drs"
    """Data Repository Service

    A data repository service is a file source that is backed by a remote data
    repository service implementing the (DRS) API which provides a generic interface
    to data repositories so data consumers, including workflow systems,
    can access data in a single, standard way regardless of where it's stored and how it's managed.
    """

    rdm = "rdm"
    """Research Data Management

    A research data management file source is a file source that is backed by a remote
    research data management system that can provide DOIs. Examples include InvenioRDM, Zenodo, etc.
    """

    stock = "stock"
    """Stock Plugin

    A stock plugin is a file source that is shipped with Galaxy and covers common
    use cases. Examples include the UserFTP, LibraryImport, UserLibraryImport, etc.
    """


class FileSourceSupports(TypedDict):
    """Feature support flags for a file source plugin"""

    # Indicates whether the file source supports pagination for listing files
    pagination: NotRequired[bool]
    # Indicates whether the file source supports server-side search for listing files
    search: NotRequired[bool]
    # Indicates whether the file source supports server-side sorting for listing files
    sorting: NotRequired[bool]


class FilesSourceProperties(TypedDict):
    """Initial set of properties used to initialize a filesource.

    Filesources can extend this typed dict to define any additional
    filesource specific properties.
    """

    file_sources_config: NotRequired[FileSourcePluginsConfig]
    id: NotRequired[str]
    label: NotRequired[str]
    doc: NotRequired[Optional[str]]
    scheme: NotRequired[str]
    writable: NotRequired[bool]
    requires_roles: NotRequired[Optional[str]]
    requires_groups: NotRequired[Optional[str]]
    disable_templating: NotRequired[Optional[bool]]
    # API helper values
    uri_root: NotRequired[str]
    type: NotRequired[str]
    browsable: NotRequired[bool]
    url: NotRequired[Optional[str]]
    supports: NotRequired[FileSourceSupports]


@dataclass
class FilesSourceOptions:
    """Options to control behavior of file source operations, such as realize_to, write_from and list."""

    # Indicates access to the FS operation with intent to write.
    # Even if a file source is "writeable" some directories (or elements) may be restricted or read-only
    # so those should be skipped while browsing with writeable=True.
    writeable: Optional[bool] = False

    # Property overrides for values initially configured through the constructor. For example
    # the HTTPFilesSource passes in additional http_headers through these properties, which
    # are merged with constructor defined http_headers. The interpretation of these properties
    # are filesystem specific.
    extra_props: Optional[FilesSourceProperties] = field(default_factory=lambda: FilesSourceProperties())


class EntryData(TypedDict):
    """Provides data to create a new entry in a file source."""

    name: str
    # May contain additional properties depending on the file source


class Entry(TypedDict):
    """Represents the result of creating a new entry in a file source."""

    name: str
    uri: str
    # May contain additional properties depending on the file source
    external_link: NotRequired[str]


class RemoteEntry(TypedDict):
    name: str
    uri: str
    path: str


TDirectoryClass = TypedDict("TDirectoryClass", {"class": Literal["Directory"]})
TFileClass = TypedDict("TFileClass", {"class": Literal["File"]})


class RemoteDirectory(RemoteEntry, TDirectoryClass):
    pass


class RemoteFile(RemoteEntry, TFileClass):
    size: int
    ctime: str


AnyRemoteEntry = Union[RemoteDirectory, RemoteFile]


class SingleFileSource(metaclass=abc.ABCMeta):
    """
    Represents a protocol handler for a single remote file that can be read by or written to by Galaxy.
    A remote file source can typically handle a url like `https://galaxyproject.org/myfile.txt` or
    `drs://myserver/123456`. The filesource abstraction allows programmatic control over the specific source
    to access, injection of credentials and access control. Filesources are typically listed and configured
    through `file_sources_conf.yml` or programmatically, as required.

    Filesources can be contextualized with a `user_context`, which contains information related to the current
    user attempting to access that filesource such as the username, preferences, roles etc., which can then
    be used by the filesource to make authorization decisions or inject credentials.

    Filesources are loaded through Galaxy's plugin system in `galaxy.util.plugin_config`.
    """

    @abc.abstractmethod
    def get_writable(self) -> bool:
        """Return a boolean indicating whether this target is writable."""

    @abc.abstractmethod
    def user_has_access(self, user_context: "OptionalUserContext") -> bool:
        """Return a boolean indicating whether the user can access the FileSource."""

    @abc.abstractmethod
    def realize_to(
        self,
        source_path: str,
        native_path: str,
        user_context: "OptionalUserContext" = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        """Realize source path (relative to uri root) to local file system path.

        :param source_path: url of the source file to copy from. e.g. `https://galaxyproject.org/myfile.txt`
        :type source_path: str
        :param native_path: local path to write to. e.g. `/tmp/myfile.txt`
        :type native_path: str
        :param user_context: A user context , defaults to None
        :type user_context: OptionalUserContext, optional
        :param opts: A set of options to exercise additional control over the realize_to method. Filesource specific, defaults to None
        :type opts: Optional[FilesSourceOptions], optional
        """

    @abc.abstractmethod
    def write_from(
        self,
        target_path: str,
        native_path: str,
        user_context: "OptionalUserContext" = None,
        opts: Optional[FilesSourceOptions] = None,
    ) -> str:
        """Write file at native path to target_path (relative to uri root).

        :param target_path: url of the target file to write to within the filesource. e.g. `gxfiles://myftp1/myfile.txt`
        :type target_path: str
        :param native_path: The local file to read. e.g. `/tmp/myfile.txt`
        :type native_path: str
        :param user_context: A user context , defaults to None
        :type user_context: _type_, optional
        :param opts: A set of options to exercise additional control over the write_from method. Filesource specific, defaults to None
        :type opts: Optional[FilesSourceOptions], optional
        :return: Actual url of the written file, fixed by the service backing the FileSource. May differ from the target
            path.
        :rtype: str
        """

    @abc.abstractmethod
    def score_url_match(self, url: str) -> int:
        """Return how well a given url matches this filesource. A score greater than zero indicates that
        this filesource is capable of processing the given url.

        Scoring is based on the following rules:
        a. The maximum score will be the length of the url.
        b. The minimum score will be the length of the scheme if the filesource can handle the file.
        c. The score will be zero if the filesource cannot handle the file.

        For example, given the following file source config:
        - type: webdav
          id: cloudstor
          url: "https://cloudstor.aarnet.edu.au"
          root: "/plus/remote.php/webdav"
        - type: http
          id: generic_http

        the generic http handler may return a score of 8 for the url
        https://cloudstor.aarnet.edu.au/plus/remote.php/webdav/myfolder/myfile.txt,
        as it can handle only the scheme part of the url. A webdav handler may return a score of
        55 for the same url, as both the webdav url and root combined are a specific match.

        :param url: The url to score for a match against this filesource.
        :type url: str
        :return: A score based on the aforementioned rules.
        :rtype: int
        """

    @abc.abstractmethod
    def to_relative_path(self, url: str) -> str:
        """Convert this url to a filesource relative path. For example, given the url
        `gxfiles://mysource1/myfile.txt` it will return `/myfile.txt`. Protocols directly understood
        by the handler need not be relativized. For example, the url `s3://bucket/myfile.txt` can be
        returned unchanged."""

    @abc.abstractmethod
    def to_dict(self, for_serialization=False, user_context: "OptionalUserContext" = None) -> FilesSourceProperties:
        """Return a dictified representation of this FileSource instance.

        If ``user_context`` is supplied, properties should be written so user
        context doesn't need to be present after the plugin is re-hydrated.
        """

    @abc.abstractmethod
    def prefer_links(self) -> bool:
        """Prefer linking to files."""


class SupportsBrowsing(metaclass=abc.ABCMeta):
    """An interface indicating that this filesource is browsable.

    Browsable filesources will typically have a root uri from which to start browsing.
    e.g. In an s3 bucket, the root uri may be gxfiles://bucket1/

    They will also have a list method to list files in a specific path within the filesource.
    """

    @abc.abstractmethod
    def get_uri_root(self) -> str:
        """Return a prefix for the root (e.g. gxfiles://prefix/)."""

    @abc.abstractmethod
    def list(
        self,
        path="/",
        recursive=False,
        user_context: "OptionalUserContext" = None,
        opts: Optional[FilesSourceOptions] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        sort_by: Optional[str] = None,
    ) -> Tuple[List[AnyRemoteEntry], int]:
        """Return a list of 'Directory's and 'File's and the total count in a tuple."""


class FilesSource(SingleFileSource, SupportsBrowsing):
    """Represents a combined interface for single or browsable filesources.
    The `get_browsable` method can be used to determine whether the filesource is browsable and
    implements the `SupportsBrowsing` interface.
    """

    plugin_type: ClassVar[str]

    @abc.abstractmethod
    def get_browsable(self) -> bool:
        """Return true if the filesource implements the SupportsBrowsing interface."""


def file_source_type_is_browsable(target_type: Type["BaseFilesSource"]) -> bool:
    # Check whether the list method has been overridden
    return target_type.list != BaseFilesSource.list or target_type._list != BaseFilesSource._list


class BaseFilesSource(FilesSource):
    plugin_kind: ClassVar[PluginKind] = PluginKind.rfs  # Remote File Source by default, override in subclasses
    supports_pagination: ClassVar[bool] = False
    supports_search: ClassVar[bool] = False
    supports_sorting: ClassVar[bool] = False

    def get_browsable(self) -> bool:
        return file_source_type_is_browsable(type(self))

    def get_prefix(self) -> Optional[str]:
        return self.id

    def get_scheme(self) -> str:
        return self.scheme or "gxfiles"

    def get_writable(self) -> bool:
        return self.writable

    def user_has_access(self, user_context: "OptionalUserContext") -> bool:
        if user_context is None and self.user_context_required:
            return False
        return (
            user_context is None
            or user_context.is_admin
            or (self._user_has_required_roles(user_context) and self._user_has_required_groups(user_context))
        )

    def prefer_links(self) -> bool:
        return False

    @property
    def user_context_required(self) -> bool:
        return self.requires_roles is not None or self.requires_groups is not None

    def get_uri_root(self) -> str:
        prefix = self.get_prefix()
        scheme = self.get_scheme()
        root = f"{scheme}://"
        if prefix:
            root = uri_join(root, prefix)
        return root

    def get_url(self) -> Optional[str]:
        """Returns a URL that can be used to link to the remote source."""
        return None

    def to_relative_path(self, url: str) -> str:
        return url.replace(self.get_uri_root(), "") or "/"

    def score_url_match(self, url: str) -> int:
        root = self.get_uri_root()
        return len(root) if root in url else 0

    def uri_from_path(self, path: str) -> str:
        uri_root = self.get_uri_root()
        return uri_join(uri_root, path)

    def _parse_common_config_opts(self, kwd: FilesSourceProperties):
        self._file_sources_config = kwd.pop("file_sources_config")
        self.id = kwd.pop("id")
        self.label = kwd.pop("label", None) or self.id
        self.doc = kwd.pop("doc", None)
        self.scheme = kwd.pop("scheme", DEFAULT_SCHEME)
        self.writable = kwd.pop("writable", DEFAULT_WRITABLE)
        self.requires_roles = kwd.pop("requires_roles", None)
        self.requires_groups = kwd.pop("requires_groups", None)
        self.disable_templating = kwd.pop("disable_templating", False)
        self._validate_security_rules()
        # If coming from to_dict, strip API helper values
        kwd.pop("uri_root", None)
        kwd.pop("type", None)
        kwd.pop("browsable", None)
        kwd.pop("supports", None)
        return kwd

    def to_dict(self, for_serialization=False, user_context: "OptionalUserContext" = None) -> FilesSourceProperties:
        rval: FilesSourceProperties = {
            "id": self.id,
            "type": self.plugin_type,
            "label": self.label,
            "doc": self.doc,
            "writable": self.writable,
            "browsable": self.get_browsable(),
            "requires_roles": self.requires_roles,
            "requires_groups": self.requires_groups,
            "disable_templating": self.disable_templating,
            "scheme": self.get_scheme(),
            "supports": {
                "pagination": self.supports_pagination,
                "search": self.supports_search,
                "sorting": self.supports_sorting,
            },
        }
        if self.get_browsable():
            rval["uri_root"] = self.get_uri_root()
        if self.get_url() is not None:
            rval["url"] = self.get_url()
        if for_serialization:
            rval.update(self._serialization_props(user_context=user_context))
        return rval

    def to_dict_time(self, ctime):
        if ctime is None:
            return None
        elif isinstance(ctime, (int, float)):
            return time.strftime("%m/%d/%Y %I:%M:%S %p", time.localtime(ctime))
        else:
            return ctime.strftime("%m/%d/%Y %I:%M:%S %p")

    @abc.abstractmethod
    def _serialization_props(self, user_context: "OptionalUserContext" = None) -> FilesSourceProperties:
        """Serialize properties needed to recover plugin configuration.
        Used in to_dict method if for_serialization is True.
        """

    def list(
        self,
        path="/",
        recursive=False,
        user_context: "OptionalUserContext" = None,
        opts: Optional[FilesSourceOptions] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        sort_by: Optional[str] = None,
    ) -> Tuple[List[AnyRemoteEntry], int]:
        self._check_user_access(user_context)
        if not self.supports_pagination and (limit is not None or offset is not None):
            raise RequestParameterInvalidException("Pagination is not supported by this file source.")
        if not self.supports_search and query:
            raise RequestParameterInvalidException("Server-side search is not supported by this file source.")
        if not self.supports_sorting and sort_by:
            raise RequestParameterInvalidException("Server-side sorting is not supported by this file source.")
        if self.supports_pagination:
            if limit is not None and limit < 1:
                raise RequestParameterInvalidException("Limit must be greater than 0.")
            if offset is not None and offset < 0:
                raise RequestParameterInvalidException("Offset must be greater than or equal to 0.")

        return self._list(path, recursive, user_context, opts, limit, offset, query)

    def _list(
        self,
        path="/",
        recursive=False,
        user_context: "OptionalUserContext" = None,
        opts: Optional[FilesSourceOptions] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        query: Optional[str] = None,
        sort_by: Optional[str] = None,
    ) -> Tuple[List[AnyRemoteEntry], int]:
        raise NotImplementedError()

    def create_entry(
        self,
        entry_data: EntryData,
        user_context: "OptionalUserContext" = None,
        opts: Optional[FilesSourceOptions] = None,
    ) -> Entry:
        self._ensure_writeable()
        self._check_user_access(user_context)
        return self._create_entry(entry_data, user_context, opts)

    def _create_entry(
        self,
        entry_data: EntryData,
        user_context: "OptionalUserContext" = None,
        opts: Optional[FilesSourceOptions] = None,
    ) -> Entry:
        """Create a new entry (directory) in the file source.

        The file source must be writeable.
        This function can be overridden by subclasses to provide a way of creating entries in the file source.
        """
        raise NotImplementedError()

    def write_from(
        self,
        target_path: str,
        native_path: str,
        user_context: "OptionalUserContext" = None,
        opts: Optional[FilesSourceOptions] = None,
    ) -> str:
        self._ensure_writeable()
        self._check_user_access(user_context)
        return self._write_from(target_path, native_path, user_context=user_context, opts=opts) or target_path

    @abc.abstractmethod
    def _write_from(
        self,
        target_path: str,
        native_path: str,
        user_context: "OptionalUserContext" = None,
        opts: Optional[FilesSourceOptions] = None,
    ) -> Optional[str]:
        pass

    def realize_to(
        self,
        source_path: str,
        native_path: str,
        user_context: "OptionalUserContext" = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        self._check_user_access(user_context)
        self._realize_to(source_path, native_path, user_context, opts=opts)

    @abc.abstractmethod
    def _realize_to(
        self,
        source_path: str,
        native_path: str,
        user_context: "OptionalUserContext" = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        pass

    def _ensure_writeable(self):
        if not self.get_writable():
            raise Exception("Cannot write to a non-writable file source.")

    def _check_user_access(self, user_context):
        """Raises an exception if the given user doesn't have the rights to access this file source.

        Warning: if the user_context is None, then the check is skipped. This is due to tool executions context
        not having access to the user_context. The validation will be done when checking the tool parameters.
        """
        if user_context is not None and not self.user_has_access(user_context):
            raise ItemAccessibilityException(f"User {user_context.username} has no access to file source.")

    def _evaluate_prop(self, prop_val: Any, user_context: "OptionalUserContext"):
        rval = prop_val

        # just return if we've disabled templating for this plugin
        if self.disable_templating:
            return rval

        if isinstance(prop_val, str) and "$" in prop_val:
            template_context = dict(
                user=user_context,
                environ=os.environ,
                config=self._file_sources_config,
            )
            rval = fill_template(prop_val, context=template_context, futurized=True)
        elif isinstance(prop_val, dict):
            rval = {key: self._evaluate_prop(childprop, user_context) for key, childprop in prop_val.items()}
        elif isinstance(prop_val, list):
            rval = [self._evaluate_prop(childprop, user_context) for childprop in prop_val]

        return rval

    def _user_has_required_roles(self, user_context: "FileSourcesUserContext") -> bool:
        if self.requires_roles:
            return self._evaluate_security_rules(self.requires_roles, user_context.role_names)
        return True

    def _user_has_required_groups(self, user_context: "FileSourcesUserContext") -> bool:
        if self.requires_groups:
            return self._evaluate_security_rules(self.requires_groups, user_context.group_names)
        return True

    def _evaluate_security_rules(self, rule_expression: str, user_credentials: Set[str]) -> bool:
        token_evaluator = TokenContainedEvaluator(user_credentials)
        evaluator = BooleanExpressionEvaluator(token_evaluator)
        return evaluator.evaluate_expression(rule_expression)

    def _validate_security_rules(self) -> None:
        """Checks if the security rules defined in the plugin configuration are valid boolean expressions or raises
        a ConfigurationError exception otherwise."""

        def _get_error_msg_for(rule_name: str) -> str:
            return f"Invalid boolean expression for '{rule_name}' in {self.label} file source plugin configuration."

        if self.requires_roles and not BooleanExpressionEvaluator.is_valid_expression(self.requires_roles):
            raise ConfigurationError(_get_error_msg_for("requires_roles"))
        if self.requires_groups and not BooleanExpressionEvaluator.is_valid_expression(self.requires_groups):
            raise ConfigurationError(_get_error_msg_for("requires_groups"))


def uri_join(*args: str) -> str:
    # url_join doesn't work with non-standard scheme
    if "://" in (arg0 := args[0]):
        scheme, path = arg0.split("://", 1)
        rval = f"{scheme}://{slash_join(path, *args[1:]) if path else slash_join(*args[1:])}"
    else:
        rval = slash_join(*args)
    return rval


def slash_join(*args: str) -> str:
    # https://codereview.stackexchange.com/questions/175421/joining-strings-to-form-a-url
    return "/".join(arg.strip("/") for arg in args)



import logging
import re
from typing import (
    cast,
    Dict,
    Optional,
)

from typing_extensions import Unpack

from galaxy.files import OptionalUserContext
from . import (
    BaseFilesSource,
    FilesSourceOptions,
    FilesSourceProperties,
    PluginKind,
)
from .util import fetch_drs_to_file

log = logging.getLogger(__name__)


class DRSFilesSourceProperties(FilesSourceProperties, total=False):
    url_regex: str
    force_http: bool
    http_headers: Dict[str, str]


class DRSFilesSource(BaseFilesSource):
    plugin_type = "drs"
    plugin_kind = PluginKind.drs

    def __init__(self, **kwd: Unpack[FilesSourceProperties]):
        kwds: FilesSourceProperties = dict(
            id="_drs",
            label="DRS file",
            doc="DRS file handler",
            writable=False,
        )
        kwds.update(kwd)
        props: DRSFilesSourceProperties = cast(DRSFilesSourceProperties, self._parse_common_config_opts(kwds))
        self._url_regex_str = props.pop("url_regex", r"^drs://")
        assert self._url_regex_str
        self._url_regex = re.compile(self._url_regex_str)
        self._force_http = props.pop("force_http", False)
        self._props = props

    @property
    def _allowlist(self):
        return self._file_sources_config.fetch_url_allowlist

    def _realize_to(
        self,
        source_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        props = self._serialization_props(user_context)
        headers = props.pop("http_headers", {}) or {}
        fetch_drs_to_file(
            source_path,
            native_path,
            user_context,
            fetch_url_allowlist=self._allowlist,
            headers=headers,
            force_http=self._force_http,
        )

    def _write_from(
        self,
        target_path: str,
        native_path: str,
        user_context: OptionalUserContext = None,
        opts: Optional[FilesSourceOptions] = None,
    ):
        raise NotImplementedError()

    def score_url_match(self, url: str):
        if match := self._url_regex.match(url):
            return match.span()[1]
        else:
            return 0

    def _serialization_props(self, user_context: OptionalUserContext = None) -> DRSFilesSourceProperties:
        effective_props = {}
        for key, val in self._props.items():
            effective_props[key] = self._evaluate_prop(val, user_context=user_context)
        effective_props["url_regex"] = self._url_regex_str
        effective_props["force_http"] = self._force_http
        return cast(DRSFilesSourceProperties, effective_props)


__all__ = ("DRSFilesSource",)



from typing import Optional

from fs.osfs import OSFS

from . import FilesSourceOptions
from ._pyfilesystem2 import PyFilesystem2FilesSource


class TempFilesSource(PyFilesystem2FilesSource):
    """A FilesSource plugin for temporary file systems.

    Used for testing and other temporary file system needs.

    Note: This plugin is not intended for production use.
    """

    plugin_type = "temp"
    required_module = OSFS

    def _open_fs(self, user_context=None, opts: Optional[FilesSourceOptions] = None):
        props = self._serialization_props(user_context)
        extra_props = opts.extra_props or {} if opts else {}
        # We use OSFS here because using TempFS or MemoryFS would wipe out the files
        # every time we instantiate a new handle, which happens on every request.
        handle = OSFS(**{**props, **extra_props})
        return handle

    def get_scheme(self) -> str:
        return "temp"


__all__ = ("TempFilesSource",)
